/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%             DDDD   IIIII  SSSSS  PPPP   L       AAA   Y   Y                 %
%             D   D    I    SS     P   P  L      A   A   Y Y                  %
%             D   D    I     SSS   PPPP   L      AAAAA    Y                   %
%             D   D    I       SS  P      L      A   A    Y                   %
%             DDDD   IIIII  SSSSS  P      LLLLL  A   A    Y                   %
%                                                                             %
%                                                                             %
%          Display Machine Independent File Format Image via X11.             %
%                                                                             %
%                                                                             %
%                                                                             %
%                           Software Design                                   %
%                             John Cristy                                     %
%                              July 1992                                      %
%                                                                             %
%                                                                             %
%  Copyright 1994 E. I. du Pont de Nemours & Company                          %
%                                                                             %
%  Permission to use, copy, modify, distribute, and sell this software and    %
%  its documentation for any purpose is hereby granted without fee,           %
%  provided that the above Copyright notice appear in all copies and that     %
%  both that Copyright notice and this permission notice appear in            %
%  supporting documentation, and that the name of E. I. du Pont de Nemours    %
%  & Company not be used in advertising or publicity pertaining to            %
%  distribution of the software without specific, written prior               %
%  permission.  E. I. du Pont de Nemours & Company makes no representations   %
%  about the suitability of this software for any purpose.  It is provided    %
%  "as is" without express or implied warranty.                               %
%                                                                             %
%  E. I. du Pont de Nemours & Company disclaims all warranties with regard    %
%  to this software, including all implied warranties of merchantability      %
%  and fitness, in no event shall E. I. du Pont de Nemours & Company be       %
%  liable for any special, indirect or consequential damages or any           %
%  damages whatsoever resulting from loss of use, data or profits, whether    %
%  in an action of contract, negligence or other tortuous action, arising     %
%  out of or in connection with the use or performance of this software.      %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Display is a machine architecture independent image processing
%  and display program.  It can display any image in the MIFF format on
%  any workstation display running X.  Display first determines the
%  hardware capabilities of the workstation.  If the number of unique
%  colors in the image is less than or equal to the number the workstation
%  can support, the image is displayed in an X window.  Otherwise the
%  number of colors in the image is first reduced to match the color
%  resolution of the workstation before it is displayed.
%
%  This means that a continuous-tone 24 bits-per-pixel image can display on a
%  8 bit pseudo-color device or monochrome device.  In most instances the
%  reduced color image closely resembles the original.  Alternatively, a
%  monochrome or pseudo-color image can display on a continuous-tone 24
%  bits-per-pixel device.
%
%  The Display program command syntax is:
%
%  Usage: display [options ...] file [ [options ...] file ...]
%
%  Where options include:
%    -alpha              store alpha channel if the image has one
%    -backdrop           display image centered on a backdrop
%    -border geometry    surround image with a border of color
%    -clip geometry      preferred size and location of the clipped image
%    -colormap type      Shared or Private
%    -colors value       preferred number of colors in the image
%    -colorspace type    GRAY, OHTA, RGB, XYZ, YCbCr, YIQ, or YUV
%    -comment string     annotate image with comment",
%    -compress type      RunlengthEncoded or QEncoded
%    -delay seconds      display the next image after pausing
%    -density geometry   vertical and horizontal density of the image
%    -display server     display image to this X server
%    -dither             apply Floyd/Steinberg error diffusion to image
%    -enhance            apply a digital filter to enhance a noisy image
%    -gamma value        level of gamma correction
%    -geometry geometry  preferred size and location of the image window
%    -interlace type     NONE, LINE, or PLANE
%    -inverse            apply color inversion to image
%    -label name         assign a label to an image
%    -map type           display image using this Standard Colormap
%    -monochrome         transform image to black and white
%    -noise              reduce noise with a noise peak elimination filter
%    -normalize          transform image to span the full the range of colors
%    -page geometry      size and location of the Postscript page
%    -quality value      JPEG quality setting
%    -reflect            reverse image scanlines
%    -roll geometry      roll an image vertically or horizontally
%    -rotate degrees     apply Paeth rotation to the image
%    -scene value        image scene number
%    -shear geometry     slide one edge of the image along the X or Y axis
%    -treedepth value    depth of the color classification tree
%    -update seconds     detect when image file is modified and redisplay
%    -verbose            print detailed information about the image
%    -visual type        display image using this visual type
%    -window id          display image to background of this window
%    -write filename     write image to a file
%
%  In addition to those listed above, you can specify these standard X
%  resources as command line options:  -background, -bordercolor,
%  -borderwidth, -font, -foreground, -iconGeometry, -iconic, -matte,
%  -name, or -title.
%
%  Change '-' to '+' in any option above to reverse its effect.  For
%  example, specify +alpha to store the image without its alpha channel.
%
%  By default, the image format of `file' is determined by its magic
%  number.  To specify a particular image format, precede the filename
%  with an image format name and a colon (i.e. ps:image) or specify the
%  image type as the filename suffix (i.e. image.ps).  Specify 'file' as
%  '-' for standard input or output.
%
%  Buttons:
%    1    press and drag to select a command from a pop-up menu
%    2    press and drag to define a region of the image to magnify
%    3    press to load an image from a montage
%
%  Keyboard accelerators:
%    i    display information about the image
%    <    half the image size
%    =    resize the image
%    >    double the image size
%    c    clip the image
%    /    rotate the image 90 degrees clockwise
%    |    rotate the image
%    \    rotate the image 90 degrees counter-clockwise
%    r    reflect the image scanlines
%    g    gamma correct the image
%    a    annotate the image with text
%    x    composite image with another
%    o    restore the image to its original size
%    p    print the image to a Postscript printer
%    w    write the image to a file
%    l    load an image from a file
%    n    display the next image
%    f    display the former image
%    q    discard all images and exit program
%    1-9  change the level of magnification
%
%
*/

/*
  Include declarations.
*/
#include "magick.h"
#include "image.h"
#include "utility.h"
#include "X.h"
#include "widget.h"

/*
  State declarations.
*/
#define ControlState  0x0001
#define DefaultState  0x0000
#define EscapeState  0x0002
#define ExitState  0x0004
#define FormerImageState  0x0008
#define NextImageState  0x0010
#define UpdateColormapState  0x0020
#define UpdateConfigurationState  0x0040

/*
  Forward declarations.
*/
static Image
  *XImageWindowCommand _Declare((Display *,XResourceInfo *,XWindows *,
    KeySym,Image **,unsigned long *)),
  *XLoadImageWindow _Declare((Display *,XResourceInfo *,XWindows *,Image *)),
  *XTileImageWindow _Declare((Display *,XResourceInfo *,XWindows *,Image *,
    XEvent *));

static unsigned int
  XPrintImageWindow _Declare((Display *,XResourceInfo *,XWindows *,Image *)),
  XReflectImageWindow _Declare((Display *,XWindows *,Image **)),
  XRotateImageWindow _Declare((Display *,XWindows *,double,Image **)),
  XWriteImageWindow _Declare((Display *,XResourceInfo *,XWindows *,Image *));

static void
  XDrawPanRectangle _Declare((Display *,XWindows *)),
  XMagnifyImageWindow _Declare((Display *,XResourceInfo *,XWindows *,XEvent *)),
  XMakePanImage _Declare((Display *,XResourceInfo *,XWindows *,Image *)),
  XPanImageWindow _Declare((Display *,XWindows *,XEvent *)),
  XMagnifyWindowCommand _Declare((Display *,XResourceInfo *,XWindows *,KeySym)),
  XSetClipGeometry _Declare((Display *,XWindows *,RectangleInfo *,Image *)),
  XTranslateImageWindow _Declare((Display *,XWindows *,Image *,KeySym));

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   U s a g e                                                                 %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function Usage displays the program command syntax.
%
%  The format of the Usage routine is:
%
%      Usage(terminate)
%
%  A description of each parameter follows:
%
%    o terminate: A value other than zero is returned if the program is to
%      terminate immediately.
%
*/
static void Usage(terminate)
unsigned int
  terminate;
{
  char
    **p;

  static char
    *buttons[]=
    {
      "1    press and drag to select a command from a pop-up menu",
      "2    press and drag to define a region of the image to magnify",
      "3    press to load an image from a montage",
      (char *) NULL
    },
    *keys[]=
    {
      "i    display information about the image",
      "<    half the image size",
      "=    resize the image",
      ">    double the image size",
      "c    clip the image",
      "/    rotate the image 90 degrees clockwise",
      "|    rotate the image",
      "\\    rotate the image 90 degrees counter-clockwise",
      "r    reflect the image scanlines",
      "g    gamma correct the image",
      "a    annotate the image with text",
      "x    composite image with another",
      "p    print the image to a Postscript printer",
      "w    write the image to a file",
      "o    restore the image to its original size",
      "l    load an image from a file",
      "n    display the next image",
      "f    display the former image",
      "q    discard all images and exit program",
      "1-9  change the level of magnification",
      (char *) NULL
    },
    *options[]=
    {
      "-alpha              store alpha channel if the image has one",
      "-backdrop           display image centered on a backdrop",
      "-border geometry    surround image with a border of color",
      "-clip geometry      preferred size and location of the clipped image",
      "-colormap type      Shared or Private",
      "-colors value       preferred number of colors in the image",
      "-colorspace type    GRAY, OHTA, RGB, XYZ, YCbCr, YIQ, or YUV",
      "-comment string     annotate image with comment",
      "-compress type      RunlengthEncoded or QEncoded",
      "-delay seconds      display the next image after pausing",
      "-density geometry   vertical and horizontal density of the image",
      "-display server     display image to this X server",
      "-dither             apply Floyd/Steinberg error diffusion to image",
      "-enhance            apply a digital filter to enhance a noisy image",
      "-gamma value        level of gamma correction",
      "-geometry geometry  preferred size and location of the image window",
      "-interlace type     NONE, LINE, or PLANE",
      "-inverse            apply color inversion to image",
      "-label name         assign a label to an image",
      "-map type           display image using this Standard Colormap",
      "-monochrome         transform image to black and white",
      "-noise              reduce noise with a noise peak elimination filter",
      "-normalize          transform image to span the full range of colors",
      "-page geometry      size and location of the Postscript page",
      "-quality value      JPEG quality setting",
      "-reflect            reflect the image scanlines",
      "-roll geometry      roll an image vertically or horizontally",
      "-rotate degrees     apply Paeth rotation to the image",
      "-scene value        image scene number",
      "-shear geometry     slide one edge of the image along the X or Y axis",
      "-treedepth value    depth of the color classification tree",
      "-update seconds     detect when image file is modified and redisplay",
      "-verbose            print detailed information about the image",
      "-visual type        display image using this visual type",
      "-window id          display image to background of this window",
      "-write filename     write image to a file",
      (char *) NULL
    };
  (void) fprintf(stderr,
    "Usage: %s [-options ...] file [ [-options ...] file ...]\n",client_name);
  (void) fprintf(stderr,"\nWhere options include: \n");
  for (p=options; *p != (char *) NULL; p++)
    (void) fprintf(stderr,"  %s\n",*p);
  (void) fprintf(stderr,
    "\nIn addition to those listed above, you can specify these standard X\n");
  (void) fprintf(stderr,
    "resources as command line options:  -background, -bordercolor,\n");
  (void) fprintf(stderr,
    "-borderwidth, -font, -foreground, -iconGeometry, -iconic, -matte, \n");
  (void) fprintf(stderr,"-name, or -title.\n");
  (void) fprintf(stderr,
    "\nChange '-' to '+' in any option above to reverse its effect.  For\n");
  (void) fprintf(stderr,
    "example, specify +alpha to store the image without an alpha channel.\n");
  (void) fprintf(stderr,
    "\nBy default, the image format of `file' is determined by its magic\n");
  (void) fprintf(stderr,
    "number.  To specify a particular image format, precede the filename\n");
  (void) fprintf(stderr,
    "with an image format name and a colon (i.e. ps:image) or specify the\n");
  (void) fprintf(stderr,
    "image type as the filename suffix (i.e. image.ps).  Specify 'file' as\n");
  (void) fprintf(stderr,"'-' for standard input or output.\n");
  (void) fprintf(stderr,"\nButtons: \n");
  for (p=buttons; *p != (char *) NULL; p++)
    (void) fprintf(stderr,"  %s\n",*p);
  (void) fprintf(stderr,"\nKeyboard accelerators: \n");
  for (p=keys; *p != (char *) NULL; p++)
    (void) fprintf(stderr,"  %s\n",*p);
  if (terminate)
    exit(1);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X A n n o t a t e I m a g e W i n d o w                                   %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XAnnotateImageWindow annotates the image with text.
%
%  The format of the XAnnotateImageWindow routine is:
%
%    XAnnotateImageWindow(display,resource_info,window,image)
%
%  A description of each parameter follows:
%
%    o display: Specifies a connection to an X server;  returned from
%      XOpenDisplay.
%
%    o resource_info: Specifies a pointer to a X11 XResourceInfo structure.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o image: Specifies a pointer to a Image structure; returned from
%      ReadImage.
%
*/
static unsigned int XAnnotateImageWindow(display,resource_info,window,image)
Display
  *display;

XResourceInfo
  *resource_info;

XWindows
  *window;

Image
  *image;
{
  char
    *p,
    text[MaxTextLength];

  Cursor
    cursor;

  GC
    annotate_context;

  int
    i,
    x,
    y;

  static unsigned int
    font_id=0,
    pen_id=0;

  unsigned int
    height,
    mask,
    status,
    width;

  unsigned long
    state,
    x_factor,
    y_factor;

  Window
    xwindow;

  XAnnotateInfo
    *annotate_info,
    *previous_info;

  XFontStruct
    *font_info;

  XEvent
    event,
    text_event;

  /*
    Map info window.
  */
  state=DefaultState;
  (void) sprintf(text," +%u+%u  ",window->image.width,window->image.height);
  XSetWindowExtents(display,&window->info,text);
  XMapWindow(display,window->info.id);
  window->info.mapped=True;
  /*
    Track pointer until button 1 is pressed.
  */
  XQueryPointer(display,window->image.id,&xwindow,&xwindow,&i,&i,&x,&y,&mask);
  XSelectInput(display,window->image.id,window->image.attributes.event_mask |
    PointerMotionMask);
  do
  {
    if (window->info.mapped)
      {
        /*
          Display pointer position.
        */
        (void) sprintf(text," %+d%+d ",x-window->image.x,y-window->image.y);
        XDisplayInfoString(display,&window->info,text);
      }
    /*
      Wait for next event.
    */
    XMaskEvent(display,ButtonPressMask | ButtonReleaseMask | ExposureMask |
      KeyPressMask | PointerMotionMask,&event);
    switch (event.type)
    {
      case ButtonPress:
      {
        if ((event.xbutton.button == Button3) &&
            (event.xbutton.state & Mod1Mask))
          {
            /*
              Convert Alt-Button3 to Button2.
            */
            event.xbutton.button=Button2;
            event.xbutton.state&=(~Mod1Mask);
          }
        switch (event.xbutton.button)
        {
          case Button1:
          {
            /*
              Change to text entering mode.
            */
            x=event.xbutton.x;
            y=event.xbutton.y;
            state|=ExitState;
            break;
          }
          case Button2:
          {
            char
              font_name[MaxTextLength],
              *MenuSelections[MaxNumberFonts];

            int
              font_number;

            /*
              Initialize menu selections.
            */
            for (i=0; i < MaxNumberFonts; i++)
              MenuSelections[i]=resource_info->font_name[i];
            MenuSelections[MaxNumberFonts-1]="Browser...";
            /*
              Select a font name from the pop-up menu.
            */
            font_number=XMenuWidget(display,window,event.xbutton.x_root,
              event.xbutton.y_root,"Font Name",MenuSelections,MaxNumberFonts,
              font_name);
            if (font_number == (MaxNumberFonts-1))
              {
                static char
                  font_name[MaxTextLength]="fixed";

                /*
                  Select a font name from a browser.
                */
                resource_info->font_name[MaxNumberFonts-1]=font_name;
                XFontBrowserWidget(display,resource_info,window,"Select",
                  font_name);
                if (*font_name == '\0')
                  break;
              }
            if (font_number >= 0)
              {
                /*
                  Initialize font info.
                */
                (void) strcpy(font_name,resource_info->font_name[font_number]);
                font_info=XLoadQueryFont(display,font_name);
                if (font_info == (XFontStruct *) NULL)
                  XNoticeWidget(display,resource_info,window,
                    "Unable to load font",font_name);
                else
                  {
                    font_id=font_number;
                    XFreeFont(display,font_info);
                  }
              }
            break;
          }
          case Button3:
          {
            char
              color_name[MaxTextLength],
              *MenuSelections[MaxNumberPens];

            int
              pen_number;

            /*
              Initialize menu selections.
            */
            for (i=0; i < MaxNumberPens; i++)
              MenuSelections[i]=resource_info->pen_color[i];
            MenuSelections[MaxNumberPens-1]="Browser...";
            /*
              Select a pen color from the pop-up menu.
            */
            pen_number=XMenuWidget(display,window,event.xbutton.x_root,
              event.xbutton.y_root,"Font Color",MenuSelections,MaxNumberPens,
              color_name);
            if (pen_number == (MaxNumberPens-1))
              {
                static char
                  color_name[MaxTextLength]="gray";

                /*
                  Select a pen color from a dialog.
                */
                resource_info->pen_color[MaxNumberPens-1]=color_name;
                XColorBrowserWidget(display,resource_info,window,"Select",
                  color_name);
                if (*color_name == '\0')
                  break;
              }
            if (pen_number >= 0)
              {
                /*
                  Set pen color.
                */
                pen_id=pen_number;
                (void) strcpy(color_name,resource_info->pen_color[pen_id]);
                status=XParseColor(display,window->image.map_info->colormap,
                  color_name,&window->image.pixel_info->pen_color[pen_id]);
                if (status == 0)
                  XNoticeWidget(display,resource_info,window,
                    "Color is not known to X server",color_name);
                (void) XAllocColor(display,window->image.map_info->colormap,
                  &window->image.pixel_info->pen_color[pen_id]);
              }
            break;
          }
        }
        break;
      }
      case Expose:
      {
        /*
          Refresh image window.
        */
        XRefreshWindow(display,&window->image,&event);
        break;
      }
      case KeyPress:
      {
        static char
          command[MaxTextLength];

        static KeySym
          key_symbol;

        /*
          Respond to a user key press.
        */
        (void) XLookupString((XKeyEvent *) &event.xkey,command,sizeof(command),
          &key_symbol,(XComposeStatus *) NULL);
        if (key_symbol == XK_Escape)
          {
            /*
              Prematurely exit.
            */
            state|=EscapeState;
            state|=ExitState;
            break;
          }
        break;
      }
      case MotionNotify:
      {
        /*
          Discard pending pointer motion events.
        */
        while (XCheckMaskEvent(display,PointerMotionMask,&event));
        x=event.xmotion.x;
        y=event.xmotion.y;
        /*
          Map and unmap info window as cursor crosses its boundaries.
        */
        if (window->info.mapped)
          {
            if ((x < (window->info.x+window->info.width)) &&
                (y < (window->info.y+window->info.height)))
              {
                XWithdrawWindow(display,window->info.id,window->info.screen);
                window->info.mapped=False;
              }
          }
        else
          if ((x > (window->info.x+window->info.width)) ||
              (y > (window->info.y+window->info.height)))
            {
              XMapWindow(display,window->info.id);
              window->info.mapped=True;
            }
        break;
      }
      default:
        break;
    }
  } while (!(state & ExitState));
  XSelectInput(display,window->image.id,window->image.attributes.event_mask);
  if (window->info.mapped)
    XWithdrawWindow(display,window->info.id,window->info.screen);
  XFlush(display);
  if (state & EscapeState)
    return(True);
  /*
    Set font info and check boundary conditions.
  */
  font_info=XLoadQueryFont(display,resource_info->font_name[font_id]);
  if (font_info == (XFontStruct *) NULL)
    {
      XNoticeWidget(display,resource_info,window,"Unable to load font",
        resource_info->font_name[font_id]);
      font_info=window->image.font_info;
    }
  if ((x+font_info->max_bounds.width) >= window->image.width)
    x=window->image.width-font_info->max_bounds.width;
  if (y < (font_info->ascent+font_info->descent))
    y=font_info->ascent+font_info->descent;
  if ((font_info->max_bounds.width > window->image.width) ||
      ((font_info->ascent+font_info->descent) >= window->image.height))
    return(False);
  /*
    Initialize annotate structure.
  */
  annotate_info=(XAnnotateInfo *) malloc(sizeof(XAnnotateInfo));
  if (annotate_info == (XAnnotateInfo *) NULL)
    return(False);
  XGetAnnotateInfo(annotate_info);
  annotate_info->x=x;
  annotate_info->y=y;
  annotate_info->height=font_info->ascent+font_info->descent;
  annotate_info->font_info=font_info;
  annotate_info->text=(char *) malloc(
    (window->image.width/Max(font_info->min_bounds.width,1)+2)*sizeof(char));
  if (annotate_info->text == (char *) NULL)
    return(False);
  /*
    Create cursor and set graphic context.
  */
  cursor=XCreateFontCursor(display,XC_pencil);
  XDefineCursor(display,window->image.id,cursor);
  annotate_context=window->image.annotate_context;
  XSetFont(display,annotate_context,font_info->fid);
  XSetForeground(display,annotate_context,
    window->image.pixel_info->pen_color[pen_id].pixel);
  /*
    Begin annotating the image with text.
  */
  state=DefaultState;
  XDrawString(display,window->image.id,annotate_context,x,y,"_",1);
  text_event.xexpose.width=(unsigned int) font_info->max_bounds.width;
  text_event.xexpose.height=font_info->max_bounds.ascent+
    font_info->max_bounds.descent;
  p=annotate_info->text;
  do
  {
    /*
      Display text cursor.
    */
    *p='\0';
    XDrawString(display,window->image.id,annotate_context,x,y,"_",1);
    /*
      Wait for next event.
    */
    XNextEvent(display,&event);
    /*
      Erase text cursor.
    */
    text_event.xexpose.x=x;
    text_event.xexpose.y=y-font_info->max_bounds.ascent;
    XRefreshWindow(display,&window->image,&text_event);
    switch (event.type)
    {
      case ButtonPress:
      {
        if (event.xbutton.button != Button2)
          break;
        /*
          Request primary selection.
        */
        XConvertSelection(display,XA_PRIMARY,XA_STRING,XA_STRING,
          window->image.id,CurrentTime);
        break;
      }
      case Expose:
      {
        if (event.xexpose.count == 0)
          {
            XAnnotateInfo
              *text_info;

            /*
              Refresh image window.
            */
            XRefreshWindow(display,&window->image,(XEvent *) NULL);
            text_info=annotate_info;
            while (text_info != (XAnnotateInfo *) NULL)
            {
              XDrawString(display,window->image.id,annotate_context,
                text_info->x,text_info->y,text_info->text,
                strlen(text_info->text));
              text_info=text_info->previous;
            }
            XDrawString(display,window->image.id,annotate_context,x,y,"_",1);
          }
        break;
      }
      case KeyPress:
      {
        static char
          command[MaxTextLength];

        static int
          length;

        static KeySym
          key_symbol;

        /*
          Respond to a user key press.
        */
        length=XLookupString((XKeyEvent *) &event.xkey,command,sizeof(command),
          &key_symbol,(XComposeStatus *) NULL);
        *(command+length)='\0';
        if (key_symbol == XK_Control_L)
          {
            state|=ControlState;
            break;
          }
        if (state & ControlState)
          switch (key_symbol)
          {
            case XK_u:
            case XK_U:
            {
              key_symbol=XK_Delete;
              break;
            }
            default:
              break;
          }
        switch (key_symbol)
        {
          case XK_BackSpace:
          {
            /*
              Erase one character.
            */
            if (p == annotate_info->text)
              if (annotate_info->previous == (XAnnotateInfo *) NULL)
                break;
              else
                {
                  /*
                    Go to end of the previous line of text.
                  */
                  annotate_info=annotate_info->previous;
                  p=annotate_info->text;
                  x=annotate_info->x+annotate_info->width;
                  y=annotate_info->y;
                  if (annotate_info->width != 0)
                    p+=strlen(annotate_info->text);
                  break;
                }
            p--;
            x-=XTextWidth(font_info,p,1);
            text_event.xexpose.x=x;
            text_event.xexpose.y=y-font_info->max_bounds.ascent;
            XRefreshWindow(display,&window->image,&text_event);
            break;
          }
          case XK_Delete:
          {
            /*
              Erase the entire line of text.
            */
            while (p != annotate_info->text)
            {
              p--;
              x-=XTextWidth(font_info,p,1);
              text_event.xexpose.x=x;
              XRefreshWindow(display,&window->image,&text_event);
            }
            break;
          }
          case XK_Escape:
          {
            /*
              Finished annotating.
            */
            annotate_info->width=XTextWidth(font_info,annotate_info->text,
              strlen(annotate_info->text));
            XRefreshWindow(display,&window->image,&text_event);
            state|=ExitState;
            break;
          }
          default:
          {
            /*
              Draw a single character on the image window.
            */
            if (state & ControlState)
              break;
            if (*command == '\0')
              break;
            *p=(*command);
            XDrawString(display,window->image.id,annotate_context,x,y,p,1);
            x+=XTextWidth(font_info,p,1);
            p++;
            if ((x+font_info->max_bounds.width) < window->image.width)
              break;
          }
          case XK_Return:
          case XK_KP_Enter:
          {
            /*
              Advance to the next line of text.
            */
            *p='\0';
            annotate_info->width=XTextWidth(font_info,annotate_info->text,
              strlen(annotate_info->text));
            if (annotate_info->next != (XAnnotateInfo *) NULL)
              {
                /*
                  Line of text already exists.
                */
                annotate_info=annotate_info->next;
                x=annotate_info->x;
                y=annotate_info->y;
                p=annotate_info->text;
                break;
              }
            annotate_info->next=(XAnnotateInfo *) malloc(sizeof(XAnnotateInfo));
            if (annotate_info->next == (XAnnotateInfo *) NULL)
              return(False);
            *annotate_info->next=(*annotate_info);
            annotate_info->next->previous=annotate_info;
            annotate_info=annotate_info->next;
            annotate_info->text=(char *) malloc((window->image.width/
              Max(font_info->min_bounds.width,1)+2)*sizeof(char));
            if (annotate_info->text == (char *) NULL)
              return(False);
            annotate_info->y+=annotate_info->height;
            if (annotate_info->y > window->image.height)
              annotate_info->y=annotate_info->height;
            annotate_info->next=(XAnnotateInfo *) NULL;
            x=annotate_info->x;
            y=annotate_info->y;
            p=annotate_info->text;
            break;
          }
        }
        break;
      }
      case KeyRelease:
      {
        static char
          command[MaxTextLength];

        static KeySym
          key_symbol;

        /*
          Respond to a user key release.
        */
        (void) XLookupString((XKeyEvent *) &event.xkey,command,sizeof(command),
          &key_symbol,(XComposeStatus *) NULL);
        if (key_symbol == XK_Control_L)
          state&=(~ControlState);
        break;
      }
      case SelectionNotify:
      {
        Atom
          type;

        int
          format;

        unsigned char
          *data;

        unsigned long
          after,
          length;

        /*
          Obtain response from primary selection.
        */
        if (event.xselection.property == (Atom) None)
          break;
        status=XGetWindowProperty(display,event.xselection.requestor,
          event.xselection.property,0L,2047L,True,XA_STRING,&type,&format,
          &length,&after,&data);
        if ((status != Success) || (type != XA_STRING) || (format == 32) ||
            (length == 0))
          break;
        /*
          Annotate image window with primary selection.
        */
        for (i=0; i < length; i++)
        {
          if (data[i] != '\n')
            {
              /*
                Draw a single character on the image window.
              */
              *p=data[i];
              XDrawString(display,window->image.id,annotate_context,x,y,p,1);
              x+=XTextWidth(font_info,p,1);
              p++;
              if ((x+font_info->max_bounds.width) < window->image.width)
                continue;
            }
          /*
            Advance to the next line of text.
          */
          *p='\0';
          annotate_info->width=XTextWidth(font_info,annotate_info->text,
            strlen(annotate_info->text));
          if (annotate_info->next != (XAnnotateInfo *) NULL)
            {
              /*
                Line of text already exists.
              */
              annotate_info=annotate_info->next;
              x=annotate_info->x;
              y=annotate_info->y;
              p=annotate_info->text;
              continue;
            }
          annotate_info->next=(XAnnotateInfo *)
            malloc(sizeof(XAnnotateInfo));
          if (annotate_info->next == (XAnnotateInfo *) NULL)
            return(False);
          *annotate_info->next=(*annotate_info);
          annotate_info->next->previous=annotate_info;
          annotate_info=annotate_info->next;
          annotate_info->text=(char *) malloc((window->image.width/
            Max(font_info->min_bounds.width,1)+2)*sizeof(char));
          if (annotate_info->text == (char *) NULL)
            return(False);
          annotate_info->y+=annotate_info->height;
          if (annotate_info->y > window->image.height)
            annotate_info->y=annotate_info->height;
          annotate_info->next=(XAnnotateInfo *) NULL;
          x=annotate_info->x;
          y=annotate_info->y;
          p=annotate_info->text;
        }
        XFree((void *) data);
        break;
      }
      default:
        break;
    }
  } while (!(state & ExitState));
  XSetFont(display,annotate_context,window->image.font_info->fid);
  XSetForeground(display,annotate_context,
    window->image.pixel_info->foreground_color.pixel);
  XFreeCursor(display,cursor);
  /*
    Annotation is relative to image configuration.
  */
  x=0;
  y=0;
  width=image->columns;
  height=image->rows;
  if (window->image.clip_geometry != (char *) NULL)
    (void) XParseGeometry(window->image.clip_geometry,&x,&y,&width,&height);
  /*
    Initialize annotated image.
  */
  XDefineCursor(display,window->image.id,window->image.busy_cursor);
  XFlush(display);
  while (annotate_info != (XAnnotateInfo *) NULL)
  {
    if (annotate_info->width == 0)
      {
        /*
          No text on this line--  go to the next line of text.
        */
        previous_info=annotate_info->previous;
        (void) free((char *) annotate_info->text);
        (void) free((char *) annotate_info);
        annotate_info=previous_info;
        continue;
      }
    /*
      Determine foreground pixel index for font color.
    */
    window->image.pixel_info->annotate_color=
      window->image.pixel_info->pen_color[pen_id];
    window->image.pixel_info->annotate_index=0;
    if (window->image.pixel_info->colors != 0)
      for (i=0; i < window->image.pixel_info->colors; i++)
        if (window->image.pixel_info->pixels[i] ==
            window->image.pixel_info->pen_color[pen_id].pixel)
          {
            window->image.pixel_info->annotate_index=i;
            break;
          }
    /*
      Define the annotate geometry string.
    */
    x_factor=UpShift(width)/window->image.ximage->width;
    annotate_info->x+=window->image.x;
    annotate_info->x=DownShift(annotate_info->x*x_factor);
    y_factor=UpShift(height)/window->image.ximage->height;
    annotate_info->y+=(window->image.y-font_info->ascent);
    annotate_info->y=DownShift(annotate_info->y*y_factor);
    (void) sprintf(annotate_info->geometry,"%ux%u%+d%+d",
      (unsigned int) DownShift(annotate_info->width*x_factor),
      (unsigned int) DownShift(annotate_info->height*y_factor),
      annotate_info->x+x,annotate_info->y+y);
    /*
      Annotate image with text.
    */
    status=XAnnotateImage(display,window->image.pixel_info,annotate_info,
      False,image);
    if (status == 0)
      return(False);
    /*
      Free up memory.
    */
    previous_info=annotate_info->previous;
    (void) free((char *) annotate_info->text);
    (void) free((char *) annotate_info);
    annotate_info=previous_info;
  }
  XDefineCursor(display,window->image.id,window->image.cursor);
  XFreeFont(display,font_info);
  window->image.stasis=False;
  return(True);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X C l i p I m a g e W i n d o w                                           %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XClipImageWindow clips the image window.
%
%  The format of the XClipImageWindow routine is:
%
%    status=XClipImageWindow(display,resource_info,window,image)
%
%  A description of each parameter follows:
%
%    o status: Function XClipImageWindow returns True if the image is
%      clipped.  False is returned is there is a memory shortage or if the
%      image fails to be cropped.
%
%    o display: Specifies a connection to an X server; returned from
%      XOpenDisplay.
%
%    o resource_info: Specifies a pointer to a X11 XResourceInfo structure.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o image: Specifies a pointer to a Image structure.
%
%
*/
static unsigned int XClipImageWindow(display,resource_info,window,image)
Display
  *display;

XResourceInfo
  *resource_info;

XWindows
  *window;

Image
  *image;
{
  char
    text[MaxTextLength];

  Cursor
    cursor;

  int
    i,
    x,
    y;

  RectangleInfo
    clip_info;

  static unsigned int
    operator = 0;

  unsigned int
    mask;

  unsigned long
    state;

  Window
    xwindow;

  XEvent
    event;

  /*
    Map info window.
  */
  state=DefaultState;
  (void) sprintf(text," +%u+%u  ",window->image.width,window->image.height);
  XSetWindowExtents(display,&window->info,text);
  XMapWindow(display,window->info.id);
  window->info.mapped=True;
  /*
    Track pointer until button 1 is pressed.
  */
  XQueryPointer(display,window->image.id,&xwindow,&xwindow,&i,&i,&x,&y,&mask);
  clip_info.x=x;
  clip_info.y=y;
  clip_info.width=0;
  clip_info.height=0;
  cursor=XCreateFontCursor(display,XC_fleur);
  XSetFunction(display,window->image.highlight_context,GXinvert);
  XSelectInput(display,window->image.id,window->image.attributes.event_mask |
    PointerMotionMask);
  do
  {
    if (window->info.mapped)
      {
        /*
          Display pointer position.
        */
        (void) sprintf(text," %+d%+d ",x-window->image.x,y-window->image.y);
        XDisplayInfoString(display,&window->info,text);
      }
    /*
      Wait for next event.
    */
    XMaskEvent(display,ButtonPressMask | ButtonReleaseMask | ExposureMask |
      KeyPressMask | PointerMotionMask,&event);
    switch (event.type)
    {
      case ButtonPress:
      {
        if ((event.xbutton.button == Button3) &&
            (event.xbutton.state & Mod1Mask))
          {
            /*
              Convert Alt-Button3 to Button2.
            */
            event.xbutton.button=Button2;
            event.xbutton.state&=(~Mod1Mask);
          }
        if (event.xbutton.button == Button1)
          {
            /*
              Note first corner of clipping rectangle-- exit loop.
            */
            XDefineCursor(display,window->image.id,cursor);
            clip_info.x=event.xbutton.x;
            clip_info.y=event.xbutton.y;
            state|=ExitState;
            break;
          }
        if (event.xbutton.button == Button2)
          {
            char
              command[MaxTextLength];

            static char
              *ClipSelections[]=
              {
                "outside",
                "inside",
              };

            /*
              Select a command from the pop-up menu.
            */
            operator=XMenuWidget(display,window,event.xbutton.x_root,
              event.xbutton.y_root,"Operations",ClipSelections,
              sizeof(ClipSelections)/sizeof(ClipSelections[0]),command)+1;
            break;
          }
        break;
      }
      case Expose:
      {
        /*
          Refresh image window.
        */
        XRefreshWindow(display,&window->image,&event);
        break;
      }
      case KeyPress:
      {
        static char
          command[MaxTextLength];

        static KeySym
          key_symbol;

        static int
          length;

        /*
          Respond to a user key press.
        */
        length=XLookupString((XKeyEvent *) &event.xkey,command,sizeof(command),
          &key_symbol,(XComposeStatus *) NULL);
        *(command+length)='\0';
        if (key_symbol == XK_Escape)
          {
            /*
              Prematurely exit.
            */
            state|=EscapeState;
            state|=ExitState;
            break;
          }
        break;
      }
      case MotionNotify:
      {
        /*
          Discard pending pointer motion events.
        */
        while (XCheckMaskEvent(display,PointerMotionMask,&event));
        x=event.xmotion.x;
        y=event.xmotion.y;
        /*
          Map and unmap info window as text cursor crosses its boundaries.
        */
        if (window->info.mapped)
          {
            if ((x < (window->info.x+window->info.width)) &&
                (y < (window->info.y+window->info.height)))
              {
                XWithdrawWindow(display,window->info.id,window->info.screen);
                window->info.mapped=False;
              }
          }
        else
          if ((x > (window->info.x+window->info.width)) ||
              (y > (window->info.y+window->info.height)))
            {
              XMapWindow(display,window->info.id);
              window->info.mapped=True;
            }
        clip_info.x=x;
        clip_info.y=y;
        break;
      }
      default:
        break;
    }
  } while (!(state & ExitState));
  XSelectInput(display,window->image.id,window->image.attributes.event_mask);
  if (state & EscapeState)
    {
      /*
        User want to exit without cropping.
      */
      XFreeCursor(display,cursor);
      return(True);
    }
  /*
    Size rectangle as pointer moves until the mouse button is released.
  */
  state=DefaultState;
  (void) sprintf(text," %ux%u+%u+%u ",window->image.width,window->image.height,
    window->image.width,window->image.height);
  XSetWindowExtents(display,&window->info,text);
  x=clip_info.x;
  y=clip_info.y;
  clip_info.width=0;
  clip_info.height=0;
  do
  {
    if ((clip_info.width > 3) && (clip_info.height > 3))
      {
        /*
          Display info and draw clipping rectangle.
        */
        if (!window->info.mapped)
          {
            XMapWindow(display,window->info.id);
            window->info.mapped=True;
          }
        (void) sprintf(text," %ux%u%+d%+d",clip_info.width,clip_info.height,
          clip_info.x,clip_info.y);
        XDisplayInfoString(display,&window->info,text);
        XHighlightRegion(display,window->image.id,
          window->image.highlight_context,&clip_info);
      }
    else
      if (window->info.mapped)
        {
          /*
            Clipping rectangle is too small;  withdraw info window.
          */
          XWithdrawWindow(display,window->info.id,window->info.screen);
          window->info.mapped=False;
        }
    /*
      Wait for next event.
    */
    XMaskEvent(display,ButtonPressMask | ButtonMotionMask | ButtonReleaseMask |
      ExposureMask,&event);
    if ((clip_info.width > 3) && (clip_info.height > 3))
      XHighlightRegion(display,window->image.id,window->image.highlight_context,
        &clip_info);
    switch (event.type)
    {
      case ButtonPress:
        break;
      case ButtonRelease:
      {
        /*
          User has committed to clipping rectangle.
        */
        clip_info.x=event.xbutton.x;
        clip_info.y=event.xbutton.y;
        state|=ExitState;
        break;
      }
      case Expose:
      {
        /*
          Refresh image window.
        */
        XRefreshWindow(display,&window->image,&event);
        break;
      }
      case MotionNotify:
      {
        /*
          Discard pending button motion events.
        */
        while (XCheckMaskEvent(display,ButtonMotionMask,&event));
        clip_info.x=event.xmotion.x;
        clip_info.y=event.xmotion.y;
      }
      default:
        break;
    }
    /*
      Check boundary conditions.
    */
    if (clip_info.x < 0)
      clip_info.x=0;
    else
      if (clip_info.x > window->image.width)
        clip_info.x=window->image.width;
    if (clip_info.x < x)
      clip_info.width=(unsigned int) (x-clip_info.x);
    else
      {
        clip_info.width=(unsigned int) (clip_info.x-x);
        clip_info.x=x;
      }
    if (clip_info.y < 0)
      clip_info.y=0;
    else
      if (clip_info.y > window->image.height)
        clip_info.y=window->image.height;
    if (clip_info.y < y)
      clip_info.height=(unsigned int) (y-clip_info.y);
    else
      {
        clip_info.height=(unsigned int) (clip_info.y-y);
        clip_info.y=y;
      }
  } while (!(state & ExitState));
  XDefineCursor(display,window->image.id,window->image.cursor);
  XFreeCursor(display,cursor);
  state=DefaultState;
  do
  {
    if ((clip_info.width > 3) && (clip_info.height > 3))
      {
        /*
          Display info and draw clipping rectangle.
        */
        if (!window->info.mapped)
          {
            XMapWindow(display,window->info.id);
            window->info.mapped=True;
          }
        (void) sprintf(text," %ux%u%+d%+d",clip_info.width,clip_info.height,
          clip_info.x,clip_info.y);
        XDisplayInfoString(display,&window->info,text);
        XHighlightRegion(display,window->image.id,
          window->image.highlight_context,&clip_info);
      }
    else
      if (window->info.mapped)
        {
          /*
            Clipping rectangle is too small;  withdraw info window.
          */
          XWithdrawWindow(display,window->info.id,window->info.screen);
          window->info.mapped=False;
        }
    /*
      Wait for next event.
    */
    XMaskEvent(display,ExposureMask | KeyPressMask | KeyReleaseMask,&event);
    if ((clip_info.width > 3) && (clip_info.height > 3))
      XHighlightRegion(display,window->image.id,window->image.highlight_context,
        &clip_info);
    switch (event.type)
    {
      case Expose:
      {
        /*
          Refresh image window.
        */
        XRefreshWindow(display,&window->image,&event);
        break;
      }
      case KeyPress:
      {
        static char
          command[MaxTextLength];

        static KeySym
          key_symbol;

        static int
          length;

        /*
          Respond to a user key press.
        */
        length=XLookupString((XKeyEvent *) &event.xkey,command,sizeof(command),
          &key_symbol,(XComposeStatus *) NULL);
        *(command+length)='\0';
        if (event.xkey.state & ControlMask)
          state|=ControlState;
        switch (key_symbol)
        {
          case XK_Escape:
          {
            /*
              Exit without cropping.
            */
            state|=EscapeState;
            state|=ExitState;
            break;
          }
          case XK_Return:
          {
            /*
              Exit.
            */
            state|=ExitState;
            break;
          }
          case XK_Up:
          {
            if (state & ControlState)
              {
                clip_info.height--;
                break;
              }
            clip_info.y--;
            clip_info.height++;
            break;
          }
          case XK_Down:
          {
            if (state & ControlState)
              {
                clip_info.y++;
                clip_info.height--;
                break;
              }
            clip_info.height++;
            break;
          }
          case XK_Left:
          {
            if (state & ControlState)
              {
                clip_info.width--;
                break;
              }
            clip_info.x--;
            clip_info.width++;
            break;
          }
          case XK_Right:
          {
            if (state & ControlState)
              {
                clip_info.x++;
                clip_info.width--;
                break;
              }
            clip_info.width++;
            break;
          }
          deafult:
            break;
        }
        break;
      }
      case KeyRelease:
      {
        /*
          Respond to a user key release.
        */
        state&=(~ControlState);
        break;
      }
      default:
        break;
    }
    /*
      Check boundary conditions.
    */
    if (clip_info.x < 0)
      clip_info.x=0;
    else
      if (clip_info.x > window->image.width)
        clip_info.x=window->image.width;
    if ((clip_info.x+clip_info.width) > window->image.width)
      clip_info.width=window->image.width-clip_info.x;
    if (clip_info.y < 0)
      clip_info.y=0;
    else
      if (clip_info.y > window->image.height)
        clip_info.y=window->image.height;
    if ((clip_info.y+clip_info.height) > window->image.height)
      clip_info.height=window->image.height-clip_info.y;
  } while (!(state & ExitState));
  XSetFunction(display,window->image.highlight_context,GXcopy);
  if (window->info.mapped)
    XWithdrawWindow(display,window->info.id,window->info.screen);
  if (state & EscapeState)
    return(True);
  if ((clip_info.width > 3) && (clip_info.height > 3))
    if ((clip_info.width != window->image.width) ||
        (clip_info.height != window->image.height))
      {
        /*
          Reconfigure image window as defined by clipping rectangle.
        */
        XSetClipGeometry(display,window,&clip_info,image);
        window->image.window_changes.width=clip_info.width;
        window->image.window_changes.height=clip_info.height;
      }
  return(True);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X C o m p o s i t e I m a g e W i n d o w                                 %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XCompositeImageWindow requests an image name from the user, reads
%  the image and composites it with the X window image at a location the user
%  chooses with the pointer.
%
%  The format of the XCompositeImageWindow routine is:
%
%    status=XCompositeImageWindow(display,resource_info,window,image)
%
%  A description of each parameter follows:
%
%    o status: Function XCompositeImageWindow returns True if the image is
%      composited.  False is returned is there is a memory shortage or if the
%      image fails to be composited.
%
%    o display: Specifies a connection to an X server;  returned from
%      XOpenDisplay.
%
%    o resource_info: Specifies a pointer to a X11 XResourceInfo structure.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o image: Specifies a pointer to a Image structure; returned from
%      ReadImage.
%
*/
static unsigned int XCompositeImageWindow(display,resource_info,window,image)
Display
  *display;

XResourceInfo
  *resource_info;

XWindows
  *window;

Image
  **image;
{
  char
    filename[MaxTextLength],
    text[MaxTextLength];

  Cursor
    cursor;

  Image
    *composite_image;

  int
    i,
    x,
    y;

  RectangleInfo
    composite_info;

  static unsigned int
    operator = OverCompositeOp;

  unsigned int
    height,
    mask,
    width;

  unsigned long
    scale_factor,
    state;

  Window
    xwindow;

  XEvent
    event;

  /*
    Request image file name from user.
  */
  (void) strcpy(filename,"x:");
  XFileBrowserWidget(display,resource_info,window,"Composite",filename);
  if (*filename == '\0')
    return(True);
  /*
    Read image.
  */
  XDefineCursor(display,window->image.id,window->image.busy_cursor);
  XFlush(display);
  (void) strcpy(resource_info->image_info->filename,filename);
  composite_image=ReadImage(resource_info->image_info);
  XDefineCursor(display,window->image.id,window->image.cursor);
  if (composite_image == (Image *) NULL)
    {
      XNoticeWidget(display,resource_info,window,"Unable to read image",
        filename);
      return(False);
    }
  /*
    Map info window.
  */
  state=DefaultState;
  (void) sprintf(text," +%u+%u  ",window->image.width,window->image.height);
  XSetWindowExtents(display,&window->info,text);
  XMapWindow(display,window->info.id);
  window->info.mapped=True;
  /*
    Track pointer until button 1 is pressed.
  */
  XQueryPointer(display,window->image.id,&xwindow,&xwindow,&i,&i,&x,&y,&mask);
  composite_info.x=x;
  composite_info.y=y;
  composite_info.width=0;
  composite_info.height=0;
  XSelectInput(display,window->image.id,window->image.attributes.event_mask |
    PointerMotionMask);
  cursor=XCreateFontCursor(display,XC_ul_angle);
  XSetFunction(display,window->image.highlight_context,GXinvert);
  do
  {
    if (window->info.mapped)
      {
        /*
          Display pointer position.
        */
        (void) sprintf(text," %+d%+d ",x-window->image.x,y-window->image.y);
        XDisplayInfoString(display,&window->info,text);
      }
    XHighlightRegion(display,window->image.id,window->image.highlight_context,
      &composite_info);
    /*
      Wait for next event.
    */
    XMaskEvent(display,ButtonPressMask | ButtonReleaseMask | ExposureMask |
      KeyPressMask | PointerMotionMask,&event);
    XHighlightRegion(display,window->image.id,window->image.highlight_context,
      &composite_info);
    switch (event.type)
    {
      case ButtonPress:
      {
        if (resource_info->debug)
          (void) fprintf(stderr,"Button Press: 0x%lx %u +%d+%d\n",
            event.xbutton.window,event.xbutton.button,event.xbutton.x,
            event.xbutton.y);
        if ((event.xbutton.button == Button3) &&
            (event.xbutton.state & Mod1Mask))
          {
            /*
              Convert Alt-Button3 to Button2.
            */
            event.xbutton.button=Button2;
            event.xbutton.state&=(~Mod1Mask);
          }
        switch (event.xbutton.button)
        {
          case Button1:
          {
            composite_info.width=composite_image->columns;
            composite_info.height=composite_image->rows;
            XDefineCursor(display,window->image.id,cursor);
            break;
          }
          case Button2:
          {
            char
              command[MaxTextLength];

            static char
              *CompositeSelections[]=
              {
                "over",
                "in",
                "out",
                "atop",
                "xor",
                "plus",
                "minus",
                "add",
                "subtract",
                "difference",
                "replace",
              };

            /*
              Select a command from the pop-up menu.
            */
            operator=XMenuWidget(display,window,event.xbutton.x_root,
              event.xbutton.y_root,"Operations",CompositeSelections,
              sizeof(CompositeSelections)/sizeof(CompositeSelections[0]),
              command)+1;
            break;
          }
          default:
            break;
        }
        break;
      }
      case ButtonRelease:
      {
        if (resource_info->debug)
          (void) fprintf(stderr,"Button Release: 0x%lx %u +%d+%d\n",
            event.xbutton.window,event.xbutton.button,event.xbutton.x,
            event.xbutton.y);
        if (event.xbutton.button == Button1)
          if ((composite_info.width != 0) && (composite_info.height != 0))
            {
              /*
                User has selected the location of the composite image.
              */
              composite_info.x=event.xbutton.x;
              composite_info.y=event.xbutton.y;
              state|=ExitState;
            }
        break;
      }
      case Expose:
      {
        /*
          Refresh image window.
        */
        if (resource_info->debug)
          (void) fprintf(stderr,"Expose: 0x%lx %dx%d+%d+%d\n",
            event.xexpose.window,event.xexpose.width,event.xexpose.height,
            event.xexpose.x,event.xexpose.y);
        XRefreshWindow(display,&window->image,&event);
        break;
      }
      case KeyPress:
      {
        static char
          command[MaxTextLength];

        static KeySym
          key_symbol;

        static int
          length;

        /*
          Respond to a user key press.
        */
        length=XLookupString((XKeyEvent *) &event.xkey,command,sizeof(command),
          &key_symbol,(XComposeStatus *) NULL);
        *(command+length)='\0';
        if (resource_info->debug)
          (void) fprintf(stderr,"Key press: 0x%lx (%s)\n",key_symbol,command);
        if (key_symbol == XK_Escape)
          {
            /*
              Prematurely exit.
            */
            DestroyImage(composite_image);
            state|=EscapeState;
            state|=ExitState;
            break;
          }
        break;
      }
      case MotionNotify:
      {
        /*
          Discard pending pointer motion events.
        */
        while (XCheckMaskEvent(display,PointerMotionMask,&event));
        x=event.xmotion.x;
        y=event.xmotion.y;
        /*
          Map and unmap info window as text cursor crosses its boundaries.
        */
        if (window->info.mapped)
          {
            if ((x < (window->info.x+window->info.width)) &&
                (y < (window->info.y+window->info.height)))
              {
                XWithdrawWindow(display,window->info.id,window->info.screen);
                window->info.mapped=False;
              }
          }
        else
          if ((x > (window->info.x+window->info.width)) ||
              (y > (window->info.y+window->info.height)))
            {
              XMapWindow(display,window->info.id);
              window->info.mapped=True;
            }
        composite_info.x=x;
        composite_info.y=y;
        break;
      }
      default:
      {
        if (resource_info->debug)
          (void) fprintf(stderr,"Event type: %d\n",event.type);
        break;
      }
    }
  } while (!(state & ExitState));
  XSetFunction(display,window->image.highlight_context,GXcopy);
  XSelectInput(display,window->image.id,window->image.attributes.event_mask);
  XDefineCursor(display,window->image.id,window->image.cursor);
  XFreeCursor(display,cursor);
  if (window->info.mapped)
    XWithdrawWindow(display,window->info.id,window->info.screen);
  XFlush(display);
  if (state & EscapeState)
    return(True);
  /*
    Image compositing is relative to image configuration.
  */
  XDefineCursor(display,window->image.id,window->image.busy_cursor);
  XFlush(display);
  x=0;
  y=0;
  width=(*image)->columns;
  height=(*image)->rows;
  if (window->image.clip_geometry != (char *) NULL)
    (void) XParseGeometry(window->image.clip_geometry,&x,&y,&width,&height);
  scale_factor=UpShift(width)/window->image.ximage->width;
  composite_info.x+=window->image.x;
  composite_info.x=DownShift(composite_info.x*scale_factor);
  composite_info.width=DownShift(composite_info.width*scale_factor);
  scale_factor=UpShift(height)/window->image.ximage->height;
  composite_info.y+=window->image.y;
  composite_info.y=DownShift(composite_info.y*scale_factor);
  composite_info.height=DownShift(composite_info.height*scale_factor);
  if ((composite_info.width != composite_image->columns) ||
      (composite_info.height != composite_image->rows))
    {
      Image
        *scaled_image;

      /*
        Scale composite image.
      */
      scaled_image=
        ScaleImage(composite_image,composite_info.width,composite_info.height);
      DestroyImage(composite_image);
      if (scaled_image == (Image *) NULL)
        {
          XDefineCursor(display,window->image.id,window->image.cursor);
          return(False);
        }
      composite_image=scaled_image;
    }
  /*
    Composite image with X image window.
  */
  CompositeImage(*image,operator,composite_image,composite_info.x+x,
    composite_info.y+y);
  DestroyImage(composite_image);
  XDefineCursor(display,window->image.id,window->image.cursor);
  window->image.stasis=False;
  return(True);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X C o n f i g u r e I m a g e W i n d o w                                 %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XConfigureImageWindow creates a new X image.  It also notifies the
%  window manager of the new image size and configures the transient widows.
%
%  The format of the XConfigureImageWindow routine is:
%
%    status=XConfigureImageWindow(display,resource_info,window,image)
%
%  A description of each parameter follows:
%
%    o status: Function XConfigureImageWindow returns True if the window is
%      resized.  False is returned is there is a memory shortage or if the
%      window fails to resize.
%
%    o display: Specifies a connection to an X server; returned from
%      XOpenDisplay.
%
%    o resource_info: Specifies a pointer to a X11 XResourceInfo structure.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o image: Specifies a pointer to a Image structure;  returned from
%      ReadImage.
%
%
*/
static unsigned int XConfigureImageWindow(display,resource_info,window,image)
Display
  *display;

XResourceInfo
  *resource_info;

XWindows
  *window;

Image
  *image;
{
#define MaxPanSize  96

  char
    text[MaxTextLength];

  unsigned int
    height,
    mask,
    status,
    width,
    x_factor,
    y_factor;

  XWindowChanges
    window_changes;

  if ((window->image.width*window->image.height) > MinInfoSize)
    {
      /*
        Map info window.
      */
      (void) strcpy(text," Configuring image... ");
      XSetWindowExtents(display,&window->info,text);
      XMapWindow(display,window->info.id);
      XDisplayInfoString(display,&window->info,text);
    }
  /*
    Resize image to fit image window dimensions.
  */
  width=window->image.window_changes.width;
  height=window->image.window_changes.height;
  if (resource_info->debug)
    (void) fprintf(stderr,"Configure Image: %dx%d=>%ux%u\n",
      window->image.ximage->width,window->image.ximage->height,width,height);
  if ((width*height) == 0)
    return(True);
  XDefineCursor(display,window->image.id,window->image.busy_cursor);
  XFlush(display);
  x_factor=UpShift(width)/window->image.ximage->width;
  y_factor=UpShift(height)/window->image.ximage->height;
  status=XMakeImage(display,resource_info,&window->image,image,width,height);
  if (resource_info->use_pixmap)
    (void) XMakePixmap(display,resource_info,&window->image);
  /*
    Notify window manager of the new configuration.
  */
  if (width > XDisplayWidth(display,window->image.screen))
    width=XDisplayWidth(display,window->image.screen);
  window_changes.width=width;
  if (height > XDisplayHeight(display,window->image.screen))
    height=XDisplayHeight(display,window->image.screen);
  window_changes.height=height;
  mask=CWWidth | CWHeight;
  if (resource_info->backdrop)
    {
      mask|=CWX | CWY;
      window_changes.x=
        (XDisplayWidth(display,window->image.screen) >> 1)-(width >> 1);
      window_changes.y=
        (XDisplayHeight(display,window->image.screen) >> 1)-(height >> 1);
    }
  XReconfigureWMWindow(display,window->image.id,window->image.screen,mask,
    &window_changes);
  if ((x_factor == UpShift(1)) && (y_factor == UpShift(1)))
    XRefreshWindow(display,&window->image,(XEvent *) NULL);
  else
    {
      unsigned int
        scale_factor;

      /*
        Update transient configuration.
      */
      scale_factor=UpShift(MaxPanSize)/window->image.ximage->width;
      if (scale_factor > (UpShift(MaxPanSize)/window->image.ximage->height))
        scale_factor=UpShift(MaxPanSize)/window->image.ximage->height;
      window_changes.width=DownShift(window->image.ximage->width*scale_factor);
      window_changes.height=
        DownShift(window->image.ximage->height*scale_factor);
      XReconfigureWMWindow(display,window->pan.id,window->pan.screen,CWWidth |
        CWHeight,&window_changes);
      XBestIconSize(display,&window->icon,image);
      window_changes.width=window->icon.width;
      window_changes.height=window->icon.height;
      XReconfigureWMWindow(display,window->icon.id,window->icon.screen,CWWidth |
        CWHeight,&window_changes);
    }
  XSetWindowBackground(display,window->image.id,
    window->image.pixel_info->background_color.pixel);
  window->magnify.x=DownShift(x_factor*window->magnify.x);
  window->magnify.y=DownShift(y_factor*window->magnify.y);
  if (window->magnify.mapped)
    XMakeMagnifyImage(display,resource_info,window);
  window->image.x=DownShift(x_factor*window->image.x);
  window->image.y=DownShift(y_factor*window->image.y);
  window->pan.clip_geometry=window->image.clip_geometry;
  if (window->pan.mapped)
    XMakePanImage(display,resource_info,window,image);
  window->icon.clip_geometry=window->image.clip_geometry;
  if ((window->image.width*window->image.height) > MinInfoSize)
    XWithdrawWindow(display,window->info.id,window->info.screen);
  XDefineCursor(display,window->image.id,window->image.cursor);
  return(status);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X D i s p l a y B a c k g r o u n d I m a g e                             %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XDisplayBackgroundImage displays an image in the background of a
%  window.
%
%  The format of the XDisplayBackgroundImage routine is:
%
%      XDisplayBackgroundImage(display,resource_info,image)
%
%  A description of each parameter follows:
%
%    o display: Specifies a connection to an X server;  returned from
%      XOpenDisplay.
%
%    o resource_info: Specifies a pointer to a X11 XResourceInfo structure.
%
%    o image: Specifies a pointer to a Image structure; returned from
%      ReadImage.
%
%
*/
static void XDisplayBackgroundImage(display,resource_info,image)
Display
  *display;

XResourceInfo
  *resource_info;

Image
  *image;
{
  Atom
    property,
    type;

  char
    visual_type[MaxTextLength];

  int
    format;

  unsigned char
    *data;

  unsigned int
    height,
    status,
    width;

  unsigned long
    after,
    length;

  Window
    root_window;

  XGCValues
    context_values;

  XPixelInfo
    pixel_info;

  XResourceInfo
    resources;

  XStandardColormap
    *map_info;

  XVisualInfo
    *visual_info;

  XWindowAttributes
    window_attributes;

  XWindowInfo
    window_info;

  /*
    Determine target window.
  */
  window_info.id=(Window) NULL;
  window_info.ximage=(XImage *) NULL;
  window_info.pixmap=(Pixmap) NULL;
  root_window=XRootWindow(display,XDefaultScreen(display));
  if (Latin1Compare(resource_info->window_id,"root") == 0)
    window_info.id=root_window;
  else
    {
      if (isdigit(*resource_info->window_id))
        window_info.id=XWindowByID(display,root_window,
          (Window) strtol((char *) resource_info->window_id,(char **) NULL,0));
      if (window_info.id == (Window) NULL)
        window_info.id=
          XWindowByName(display,root_window,resource_info->window_id);
      if (window_info.id == (Window) NULL)
        Error("No window with specified id exists",resource_info->window_id);
    }
  /*
    Determine window visual id.
  */
  window_attributes.width=XDisplayWidth(display,XDefaultScreen(display));
  window_attributes.height=XDisplayHeight(display,XDefaultScreen(display));
  (void) strcpy(visual_type,"default");
  status=XGetWindowAttributes(display,window_info.id,&window_attributes);
  if (status != False)
    (void) sprintf(visual_type,"0x%lx",
      XVisualIDFromVisual(window_attributes.visual));
  /*
    Allocate standard colormap.
  */
  map_info=XAllocStandardColormap();
  if (map_info == (XStandardColormap *) NULL)
    Error("Unable to create standard colormap","Memory allocation failed");
  map_info->colormap=(Colormap) NULL;
  pixel_info.pixels=(unsigned long *) NULL;
  /*
    Initialize visual info.
  */
  resources=(*resource_info);
  resources.map_type=(char *) NULL;
  resources.visual_type=visual_type;
  visual_info=XBestVisualInfo(display,map_info,&resources);
  if (visual_info == (XVisualInfo *) NULL)
    Error("Unable to get visual",resource_info->visual_type);
  if (window_info.id == root_window)
    {
      /*
        If there are previous resources on the root window, destroy them.
      */
      property=XInternAtom(display,"_XSETROOT_ID",False);
      if (property == (Atom) NULL)
        Error("Unable to create X property","_XSETROOT_ID");
      (void) XGetWindowProperty(display,window_info.id,property,0L,1L,True,
        (Atom) AnyPropertyType,&type,&format,&length,&after,&data);
      if ((type == XA_PIXMAP) && (format == 32) && (length == 1) &&
          (after == 0))
        {
          /*
            Free previous resources on the window.
          */
          XKillClient(display,(XID) (*((Pixmap *) data)));
          XFree((void *) data);
        }
    }
  /*
    Initialize colormap.
  */
  resources.colormap=SharedColormap;
  XMakeStandardColormap(display,visual_info,&resources,image,map_info,
    &pixel_info);
  /*
    Graphic context superclass.
  */
  context_values.background=pixel_info.background_color.pixel;
  context_values.foreground=pixel_info.foreground_color.pixel;
  pixel_info.annotate_context=XCreateGC(display,window_info.id,GCBackground |
    GCForeground,&context_values);
  if (pixel_info.annotate_context == (GC) NULL)
    Error("Unable to create graphic context",(char *) NULL);
  /*
    Initialize image window attributes.
  */
  XGetWindowInfo(display,visual_info,map_info,&pixel_info,(XFontStruct *) NULL,
    resource_info,&window_info);
  /*
    Create the X image.
  */
  window_info.width=image->columns;
  if (window_info.width >= window_attributes.width)
    window_info.width=window_attributes.width;
  window_info.height=image->rows;
  if (window_info.height >= window_attributes.height)
    window_info.height=window_attributes.height;
  status=XMakeImage(display,resource_info,&window_info,image,window_info.width,
    window_info.height);
  if (status == False)
    Error("Unable to create X image",(char *) NULL);
  /*
    Adjust image dimensions as specified by backdrop or geometry options.
  */
  width=window_info.width;
  height=window_info.height;
  if (resource_info->backdrop)
    {
      /*
        Center image on root window.
      */
      window_info.x=(window_attributes.width >> 1)-
        (window_info.ximage->width >> 1);
      window_info.y=(window_attributes.height >> 1)-
        (window_info.ximage->height >> 1);
      width=window_attributes.width;
      height=window_attributes.height;
    }
  if (resource_info->image_geometry != (char *) NULL)
    {
      char
        default_geometry[MaxTextLength];

      int
        flags,
        gravity;

      XSizeHints
        *size_hints;

      /*
        User specified geometry.
      */
      size_hints=XAllocSizeHints();
      if (size_hints == (XSizeHints *) NULL)
        Error("Unable to display on window","Memory allocation failed");
      size_hints->flags=(long) NULL;
      (void) sprintf(default_geometry,"%ux%u",width,height);
      flags=XWMGeometry(display,visual_info->screen,
        resource_info->image_geometry,default_geometry,
        window_info.border_width,size_hints,&window_info.x,&window_info.y,
        (int *) &width,(int *) &height,&gravity);
      if (flags & (XValue | YValue))
        {
          width=window_attributes.width;
          height=window_attributes.height;
        }
      XFree((void *) size_hints);
    }
  /*
    Create the X pixmap.
  */
  window_info.pixmap=
    XCreatePixmap(display,window_info.id,width,height,window_info.depth);
  if (window_info.pixmap == (Pixmap) NULL)
    Error("Unable to create X pixmap",(char *) NULL);
  /*
    Display pixmap on the window.
  */
  if ((width > window_info.width) || (height > window_info.height))
    XFillRectangle(display,window_info.pixmap,window_info.highlight_context,
      0,0,width,height);
  XPutImage(display,window_info.pixmap,window_info.annotate_context,
    window_info.ximage,0,0,window_info.x,window_info.y,window_info.width,
    window_info.height);
  XSetWindowBackgroundPixmap(display,window_info.id,window_info.pixmap);
  XClearWindow(display,window_info.id);
  /*
    Free resources.
  */
  XFreePixmap(display,window_info.pixmap);
  XDestroyImage(window_info.ximage);
  XFreeGC(display,window_info.annotate_context);
  XFreeCursor(display,window_info.cursor);
  XFreeCursor(display,window_info.busy_cursor);
  if (pixel_info.pixels != (unsigned long *) NULL)
    (void) free((char *) pixel_info.pixels);
  XFree((void *) map_info);
  XFree((void *) visual_info);
  if (window_info.id == root_window)
    {
      /*
        Put property on root window and set close-down mode to RetainPermanent.
      */
      window_info.pixmap=XCreatePixmap(display,window_info.id,1,1,1);
      if (window_info.pixmap == (Pixmap) NULL)
        Error("Unable to create X pixmap",(char *) NULL);
      XChangeProperty(display,window_info.id,property,XA_PIXMAP,32,
        PropModeReplace,(unsigned char *) &window_info.pixmap,1);
      XSetCloseDownMode(display,RetainPermanent);
    }
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X D i s p l a y I m a g e                                                 %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XDisplayImage displays an image via X11.  A new image is created
%  and returned if the user interactively transforms the displayed image.
%
%  The format of the XDisplayImage routine is:
%
%      loaded_image=XDisplayImage(display,resource_info,argv,argc,image,state)
%
%  A description of each parameter follows:
%
%    o loaded_image:  Function XDisplayImage returns an image when the
%      user chooses 'Load Image' from the command menu or picks a tile
%      from the image directory.  Otherwise a null image is returned.
%
%    o display: Specifies a connection to an X server;  returned from
%      XOpenDisplay.
%
%    o resource_info: Specifies a pointer to a X11 XResourceInfo structure.
%
%    o argv: Specifies the application's argument list.
%
%    o argc: Specifies the number of arguments.
%
%    o image: Specifies an address to an address of an Image structure;
%      returned from ReadImage.
%
%
*/
static Image *XDisplayImage(display,resource_info,argv,argc,image,state)
Display
  *display;

XResourceInfo
  *resource_info;

char
  **argv;

int
  argc;

Image
  **image;

unsigned long
  *state;
{
#define MagnifySize  256  /* must be a power of 2 */
#define MaxWindows  9

  static unsigned char
    HighlightBitmap[] = {0xaa,0x55,0xaa,0x55,0xaa,0x55,0xaa,0x55},
    ShadowBitmap[] = {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};

  char
    command[MaxTextLength];

  Image
    *displayed_image,
    *loaded_image;

  int
    status;

  KeySym
    key_symbol;

  register int
    i;

  static Atom
    delete_property,
    protocols_property;

  static Window
    root_window;

  static XClassHint
    *class_hint;

  static XImage
    former_ximage;

  static XPixelInfo
    icon_pixel,
    pixel_info;

  static XPoint
    montage_info;

  static XResourceInfo
    icon_resources;

  static XStandardColormap
    *icon_map,
    *map_info;

  static XVisualInfo
    *icon_visual,
    *visual_info = (XVisualInfo *) NULL;

  static XWindowInfo
    *magick_windows[MaxWindows];

  static XWindows
    *window;

  static XWMHints
    *manager_hints;

  static unsigned int
    number_windows;

  struct stat
    file_info;

  time_t
    timer,
    time_stamp,
    update_time;

  unsigned int
    scale_factor;

  XEvent
    event;

  XFontStruct
    *font_info;

  XGCValues
    context_values;

  XWindowChanges
    window_changes;

  if (visual_info == (XVisualInfo *) NULL)
    {
      /*
        Allocate standard colormap.
      */
      if (resource_info->debug)
        XSynchronize(display,True);
      map_info=XAllocStandardColormap();
      icon_map=XAllocStandardColormap();
      if ((map_info == (XStandardColormap *) NULL) ||
          (icon_map == (XStandardColormap *) NULL))
        Error("Unable to create standard colormap","Memory allocation failed");
      map_info->colormap=(Colormap) NULL;
      icon_map->colormap=(Colormap) NULL;
      pixel_info.pixels=(unsigned long *) NULL;
      icon_pixel.pixels=(unsigned long *) NULL;
      /*
        Allocate visual.
      */
      icon_resources=(*resource_info);
      icon_resources.map_type=(char *) NULL;
      icon_resources.visual_type="default";
      icon_resources.colormap=SharedColormap;
      visual_info=XBestVisualInfo(display,map_info,resource_info);
      icon_visual=XBestVisualInfo(display,icon_map,&icon_resources);
      if ((visual_info == (XVisualInfo *) NULL) ||
          (icon_visual == (XVisualInfo *) NULL))
        Error("Unable to get visual",resource_info->visual_type);
      if (resource_info->debug)
        {
          (void) fprintf(stderr,"Visual:\n");
          (void) fprintf(stderr,"  visual id: 0x%lx\n",visual_info->visualid);
          (void) fprintf(stderr,"  class: %s\n",
            XVisualClassName(visual_info->class));
          (void) fprintf(stderr,"  depth: %d planes\n",visual_info->depth);
          (void) fprintf(stderr,"  size of colormap: %d entries\n",
            visual_info->colormap_size);
          (void) fprintf(stderr,"  red, green, blue masks: 0x%lx 0x%lx 0x%lx\n",
            visual_info->red_mask,visual_info->green_mask,
            visual_info->blue_mask);
          (void) fprintf(stderr,"  significant bits in color: %d bits\n",
            visual_info->bits_per_rgb);
        }
      /*
        Allocate atoms.
      */
      protocols_property=XInternAtom(display,"WM_PROTOCOLS",False);
      delete_property=XInternAtom(display,"WM_DELETE_WINDOW",False);
      if ((protocols_property == (Atom) NULL) ||
          (delete_property == (Atom) NULL))
        Error("Unable to create property",(char *) NULL);
      /*
        Allocate class and manager hints.
      */
      class_hint=XAllocClassHint();
      manager_hints=XAllocWMHints();
      if ((class_hint == (XClassHint *) NULL) ||
          (manager_hints == (XWMHints *) NULL))
        Error("Unable to allocate X hints",(char *) NULL);
      /*
        Initialize window id's.
      */
      root_window=XRootWindow(display,visual_info->screen);
      window=(XWindows *) malloc(sizeof(XWindows));
      if (window == (XWindows *) NULL)
        Error("Unable to create X windows","Memory allocation failed");
      number_windows=0;
      magick_windows[number_windows++]=(&window->backdrop);
      magick_windows[number_windows++]=(&window->icon);
      magick_windows[number_windows++]=(&window->image);
      magick_windows[number_windows++]=(&window->info);
      magick_windows[number_windows++]=(&window->magnify);
      magick_windows[number_windows++]=(&window->pan);
      magick_windows[number_windows++]=(&window->popup);
      for (i=0; i < number_windows; i++)
        magick_windows[i]->id=(Window) NULL;
      montage_info.x=0;
      montage_info.y=0;
    }
  /*
    Initialize Standard Colormap.
  */
  loaded_image=(Image *) NULL;
  displayed_image=(*image);
  if (resource_info->debug)
    {
      (void) fprintf(stderr,"Image: [%u] %s %ux%u ",displayed_image->scene,
        displayed_image->filename,displayed_image->columns,
        displayed_image->rows);
      if (displayed_image->colors != 0)
        (void) fprintf(stderr,"%uc ",displayed_image->colors);
      (void) fprintf(stderr,"%s\n",displayed_image->magick);
    }
  XMakeStandardColormap(display,visual_info,resource_info,displayed_image,
    map_info,&pixel_info);
  /*
    Initialize font info.
  */
  font_info=XBestFont(display,resource_info);
  if (font_info == (XFontStruct *) NULL)
    Error("Unable to load font",resource_info->font);
  /*
    Initialize graphic context.
  */
  window->context.id=(Window) NULL;
  XGetWindowInfo(display,visual_info,map_info,&pixel_info,font_info,
    resource_info,&window->context);
  class_hint->res_name="superclass";
  class_hint->res_class="ImageMagick";
  manager_hints->flags=InputHint | StateHint;
  manager_hints->input=False;
  manager_hints->initial_state=WithdrawnState;
  XMakeWindow(display,root_window,argv,argc,class_hint,manager_hints,
    delete_property,&window->context);
  if (resource_info->debug)
    (void) fprintf(stderr,"Window id: 0x%lx (context)\n",window->context.id);
  context_values.background=pixel_info.background_color.pixel;
  context_values.font=font_info->fid;
  context_values.foreground=pixel_info.foreground_color.pixel;
  context_values.graphics_exposures=False;
  pixel_info.annotate_context=XCreateGC(display,window->context.id,
    GCBackground | GCFont | GCForeground | GCGraphicsExposures,&context_values);
  if (pixel_info.annotate_context == (GC) NULL)
    Error("Unable to create graphic context",(char *) NULL);
  pixel_info.widget_context=XCreateGC(display,window->context.id,
    GCBackground | GCFont | GCForeground | GCGraphicsExposures,&context_values);
  if (pixel_info.widget_context == (GC) NULL)
    Error("Unable to create graphic context",(char *) NULL);
  context_values.background=pixel_info.foreground_color.pixel;
  context_values.foreground=pixel_info.background_color.pixel;
  context_values.plane_mask=
    context_values.background ^ context_values.foreground;
  pixel_info.highlight_context=XCreateGC(display,window->context.id,
    GCBackground | GCFont | GCForeground | GCGraphicsExposures | GCPlaneMask,
    &context_values);
  if (pixel_info.highlight_context == (GC) NULL)
    Error("Unable to create graphic context",(char *) NULL);
  XDestroyWindow(display,window->context.id);
  /*
    Initialize icon window.
  */
  XGetWindowInfo(display,icon_visual,icon_map,&icon_pixel,(XFontStruct *) NULL,
    &icon_resources,&window->icon);
  XBestIconSize(display,&window->icon,displayed_image);
  window->icon.attributes.colormap=
    XDefaultColormap(display,icon_visual->screen);
  window->icon.attributes.event_mask=StructureNotifyMask;
  class_hint->res_name="icon";
  manager_hints->flags=InputHint | StateHint;
  manager_hints->input=False;
  manager_hints->initial_state=IconicState;
  XMakeWindow(display,root_window,argv,argc,class_hint,manager_hints,
    delete_property,&window->icon);
  if (resource_info->debug)
    (void) fprintf(stderr,"Window id: 0x%lx (icon)\n",window->icon.id);
  /*
    Initialize graphic context for icon window.
  */
  context_values.background=icon_pixel.background_color.pixel;
  context_values.foreground=icon_pixel.foreground_color.pixel;
  icon_pixel.annotate_context=XCreateGC(display,window->icon.id,
    GCBackground | GCForeground,&context_values);
  if (icon_pixel.annotate_context == (GC) NULL)
    Error("Unable to create graphic context",(char *) NULL);
  window->icon.annotate_context=icon_pixel.annotate_context;
  /*
    Initialize image window.
  */
  XGetWindowInfo(display,visual_info,map_info,&pixel_info,font_info,
    resource_info,&window->image);
  window->image.name=(char *) malloc(MaxTextLength*sizeof(char));
  window->image.icon_name=(char *) malloc(MaxTextLength*sizeof(char));
  if ((window->image.name == NULL) || (window->image.icon_name == NULL))
    Error("Unable to create image window","Memory allocation failed");
  if (resource_info->title != (char *) NULL)
    {
      /*
        User specified window name.
      */
      (void) strcpy(window->image.name,resource_info->title);
      (void) strcpy(window->image.icon_name,resource_info->title);
    }
  else
    {
      char
        *p;

      /*
        Window name is the base of the filename.
      */
      p=displayed_image->filename+strlen(displayed_image->filename)-1;
      while ((p > displayed_image->filename) && (*(p-1) != '/'))
        p--;
      if ((displayed_image->previous == (Image *) NULL) &&
          (displayed_image->next == (Image *) NULL))
        (void) sprintf(window->image.name,"ImageMagick: %s",p);
      else
        (void) sprintf(window->image.name,"ImageMagick: %s.%u",p,
          displayed_image->scene);
      (void) strcpy(window->image.icon_name,p);
    }
  window->image.geometry=resource_info->image_geometry;
  window->image.width=displayed_image->columns;
  if (window->image.width > XDisplayWidth(display,visual_info->screen))
    window->image.width=XDisplayWidth(display,visual_info->screen);
  window->image.height=displayed_image->rows;
  if (window->image.height > XDisplayHeight(display,visual_info->screen))
    window->image.height=XDisplayHeight(display,visual_info->screen);
  window->image.attributes.event_mask=ButtonMotionMask | ButtonPressMask |
    ButtonReleaseMask | EnterWindowMask | ExposureMask | KeyPressMask |
    KeyReleaseMask | LeaveWindowMask | OwnerGrabButtonMask |
    StructureNotifyMask;
  XGetWindowInfo(display,visual_info,map_info,&pixel_info,font_info,
    resource_info,&window->backdrop);
  if (resource_info->backdrop || (window->backdrop.id != (Window) NULL))
    {
      /*
        Initialize backdrop window.
      */
      window->backdrop.name="ImageMagick Backdrop";
      window->backdrop.flags=USSize | USPosition;
      window->backdrop.width=XDisplayWidth(display,visual_info->screen);
      window->backdrop.height=XDisplayHeight(display,visual_info->screen);
      window->backdrop.border_width=0;
      window->backdrop.immutable=True;
      window->backdrop.attributes.do_not_propagate_mask=ButtonPressMask |
        ButtonReleaseMask;
      window->backdrop.attributes.event_mask=ButtonPressMask | KeyPressMask |
        OwnerGrabButtonMask;
      window->backdrop.attributes.override_redirect=True;
      class_hint->res_name="backdrop";
      manager_hints->flags=IconWindowHint | InputHint | StateHint;
      manager_hints->icon_window=window->icon.id;
      manager_hints->input=True;
      manager_hints->initial_state=
        resource_info->iconic ? IconicState : NormalState;
      XMakeWindow(display,root_window,argv,argc,class_hint,manager_hints,
        delete_property,&window->backdrop);
      if (resource_info->debug)
        (void) fprintf(stderr,"Window id: 0x%lx (backdrop)\n",
          window->backdrop.id);
      XMapWindow(display,window->backdrop.id);
      XClearWindow(display,window->backdrop.id);
      if (window->image.id != (Window) NULL)
        {
          XDestroyWindow(display,window->image.id);
          window->image.id=(Window) NULL;
        }
      /*
        Position image in the center the backdrop.
      */
      window->image.flags|=USPosition;
      window->image.x=(XDisplayWidth(display,visual_info->screen) >> 1)-
        (window->image.width >> 1);
      window->image.y=(XDisplayHeight(display,visual_info->screen) >> 1)-
        (window->image.height >> 1);
      window->image.attributes.event_mask|=FocusChangeMask;
    }
  if (resource_info->name == (char *) NULL)
    class_hint->res_name=client_name;
  else
    class_hint->res_name=resource_info->name;
  manager_hints->flags=IconWindowHint | InputHint | StateHint;
  manager_hints->icon_window=window->icon.id;
  manager_hints->input=True;
  manager_hints->initial_state=
    resource_info->iconic ? IconicState : NormalState;
  XMakeWindow(display,
    (resource_info->backdrop ? window->backdrop.id : root_window),argv,argc,
    class_hint,manager_hints,delete_property,&window->image);
  if (resource_info->debug)
    (void) fprintf(stderr,"Window id: 0x%lx (image)\n",window->image.id);
  /*
    Initialize X image structure.
  */
  window->image.x=0;
  window->image.y=0;
  status=XMakeImage(display,resource_info,&window->image,displayed_image,
    displayed_image->columns,displayed_image->rows);
  if (status == False)
    Error("Unable to create X image",(char *) NULL);
  if (resource_info->use_pixmap)
    (void) XMakePixmap(display,resource_info,&window->image);
  if (window->popup.id == (Window) NULL)
    XMapWindow(display,window->image.id);
  else
    if ((window->image.ximage->width == former_ximage.width) &&
        (window->image.ximage->height == former_ximage.height))
      XRefreshWindow(display,&window->image,(XEvent *) NULL);
  /*
    Initialize magnify window and cursor.
  */
  XGetWindowInfo(display,visual_info,map_info,&pixel_info,font_info,
    resource_info,&window->magnify);
  window->magnify.name=(char *) malloc(MaxTextLength*sizeof(char));
  if (window->magnify.name == NULL)
    Error("Unable to create magnify window","Memory allocation failed");
  (void) sprintf(window->magnify.name,"Magnify %uX",resource_info->magnify);
  window->magnify.cursor=XMakeCursor(display,window->image.id,
    map_info->colormap,resource_info->background_color,
    resource_info->foreground_color);
  if (window->magnify.cursor == (Cursor) NULL)
    Error("Unable to create cursor",(char *) NULL);
  window->magnify.width=MagnifySize;
  window->magnify.height=MagnifySize;
  window->magnify.x=XDisplayWidth(display,window->magnify.screen) >> 1;
  window->magnify.y=XDisplayHeight(display,window->magnify.screen) >> 1;
  window->magnify.min_width=MagnifySize;
  window->magnify.min_height=MagnifySize;
  window->magnify.width_inc=MagnifySize;
  window->magnify.height_inc=MagnifySize;
  window->magnify.attributes.save_under=True;
  window->magnify.attributes.cursor=window->magnify.cursor;
  window->magnify.attributes.event_mask=ButtonPressMask | ExposureMask |
    KeyPressMask | OwnerGrabButtonMask | StructureNotifyMask;
  class_hint->res_name="magnify";
  manager_hints->flags=InputHint | StateHint | WindowGroupHint;
  manager_hints->input=False;
  manager_hints->initial_state=NormalState;
  manager_hints->window_group=window->image.id;
  XMakeWindow(display,root_window,argv,argc,class_hint,manager_hints,
    delete_property,&window->magnify);
  if (resource_info->debug)
    (void) fprintf(stderr,"Window id: 0x%lx (magnify)\n",window->magnify.id);
  XSetTransientForHint(display,window->magnify.id,window->image.id);
  /*
    Initialize panning window.
  */
  XGetWindowInfo(display,visual_info,map_info,&pixel_info,font_info,
    resource_info,&window->pan);
  window->pan.name="Pan Icon";
  scale_factor=UpShift(MaxPanSize)/displayed_image->columns;
  if (scale_factor > (UpShift(MaxPanSize)/displayed_image->rows))
    scale_factor=UpShift(MaxPanSize)/displayed_image->rows;
  window->pan.width=DownShift(displayed_image->columns*scale_factor);
  window->pan.height=DownShift(displayed_image->rows*scale_factor);
  window->pan.x=XDisplayWidth(display,window->pan.screen) >> 1;
  window->pan.y=XDisplayHeight(display,window->pan.screen) >> 1;
  window->pan.immutable=True;
  window->pan.attributes.save_under=True;
  window->pan.attributes.event_mask=ButtonMotionMask | ButtonPressMask |
    ButtonReleaseMask | OwnerGrabButtonMask | KeyPressMask | KeyReleaseMask |
    StructureNotifyMask;
  class_hint->res_name="pan";
  manager_hints->flags=InputHint | StateHint | WindowGroupHint;
  manager_hints->input=False;
  manager_hints->initial_state=NormalState;
  manager_hints->window_group=window->image.id;
  XMakeWindow(display,root_window,argv,argc,class_hint,manager_hints,
    delete_property,&window->pan);
  if (resource_info->debug)
    (void) fprintf(stderr,"Window id: 0x%lx (pan)\n",window->pan.id);
  XSetTransientForHint(display,window->pan.id,window->image.id);
  if ((window->image.width < window->image.ximage->width) ||
      (window->image.height < window->image.ximage->height))
    XMapRaised(display,window->pan.id);
  /*
    Initialize popup window.
  */
  XGetWindowInfo(display,visual_info,map_info,&pixel_info,font_info,
    resource_info,&window->popup);
  window->popup.name=(char *) malloc(MaxTextLength*sizeof(char));
  if (window->popup.name == NULL)
    Error("Unable to create popup window","Memory allocation failed");
  window->popup.flags|=PPosition;
  window->popup.attributes.backing_store=WhenMapped;
  window->popup.attributes.save_under=True;
  window->popup.attributes.event_mask=ButtonMotionMask | ButtonPressMask |
    ButtonReleaseMask | EnterWindowMask | ExposureMask | KeyPressMask |
    LeaveWindowMask | OwnerGrabButtonMask | StructureNotifyMask;
  class_hint->res_name="popup";
  manager_hints->flags=InputHint | StateHint | WindowGroupHint;
  manager_hints->input=True;
  manager_hints->initial_state=NormalState;
  manager_hints->window_group=window->image.id;
  XMakeWindow(display,root_window,argv,argc,class_hint,manager_hints,
    delete_property,&window->popup);
  if (resource_info->debug)
    (void) fprintf(stderr,"Window id: 0x%lx (pop up)\n",window->popup.id);
  window->popup.highlight_stipple=XCreateBitmapFromData(display,
    window->popup.id,(char *) HighlightBitmap,8,8);
  window->popup.shadow_stipple=XCreateBitmapFromData(display,
    window->popup.id,(char *) ShadowBitmap,8,8);
  XSetTransientForHint(display,window->popup.id,window->image.id);
  /*
    Initialize info window.
  */
  XGetWindowInfo(display,visual_info,map_info,&pixel_info,font_info,
    resource_info,&window->info);
  window->info.name="ImageMagick Info";
  window->info.x=2;
  window->info.y=2;
  window->info.flags|=PPosition;
  window->info.attributes.save_under=True;
  window->info.attributes.win_gravity=UnmapGravity;
  window->info.attributes.event_mask=ButtonPressMask | OwnerGrabButtonMask |
    StructureNotifyMask;
  class_hint->res_name="info";
  manager_hints->flags=InputHint | StateHint | WindowGroupHint;
  manager_hints->input=False;
  manager_hints->initial_state=NormalState;
  manager_hints->window_group=window->image.id;
  XMakeWindow(display,window->image.id,argv,argc,class_hint,manager_hints,
    delete_property,&window->info);
  if (resource_info->debug)
    (void) fprintf(stderr,"Window id: 0x%lx (info)\n",window->info.id);
  /*
    Respond to events.
  */
  timer=time((time_t *) NULL)+resource_info->delay;
  update_time=0;
  if (resource_info->update)
    {
      /*
        Determine when file data was last modified.
      */
      status=stat(displayed_image->filename,&file_info);
      if (status == 0)
        update_time=file_info.st_mtime;
    }
  *state&=(~FormerImageState);
  *state&=(~NextImageState);
  do
  {
    /*
      Handle a window event.
    */
    if (resource_info->delay != 0)
      {
        if (timer < time((time_t *) NULL))
          if (!resource_info->update)
            *state|=NextImageState | ExitState;
          else
            {
              /*
                Determine if image file was modified.
              */
              status=stat(displayed_image->filename,&file_info);
              if (status == 0)
                if (update_time != file_info.st_mtime)
                  {
                    /*
                      Redisplay image.
                    */
                    loaded_image=ReadImage(resource_info->image_info);
                    if (loaded_image != (Image *) NULL)
                      *state|=NextImageState | ExitState;
                  }
              timer=time((time_t *) NULL)+resource_info->delay;
            }
        if (XEventsQueued(display,QueuedAfterFlush) == 0)
          {
            /*
              Do not block if delay > 0.
            */
            XDelay(display,SuspendTime << 2);
            continue;
          }
      }
    time_stamp=time((time_t *) NULL);
    XNextEvent(display,&event);
    if (!window->image.stasis)
      window->image.stasis=(time((time_t *) NULL)-time_stamp) > 0;
    switch (event.type)
    {
      case ButtonPress:
      {
        if (resource_info->debug)
          (void) fprintf(stderr,"Button Press: 0x%lx %u +%d+%d\n",
            event.xbutton.window,event.xbutton.button,event.xbutton.x,
            event.xbutton.y);
        if ((event.xbutton.button == Button3) &&
            (event.xbutton.state & Mod1Mask))
          {
            /*
              Convert Alt-Button3 to Button2.
            */
            event.xbutton.button=Button2;
            event.xbutton.state&=(~Mod1Mask);
          }
        if (event.xbutton.window == window->image.id)
          switch (event.xbutton.button)
          {
            case Button1:
            {
              int
                command_number;

              static char
                *MenuSelections[]=
                {
                  "Image Info",
                  "Resize",
                  "Clip",
                  "Rotate",
                  "Reflect",
                  "Gamma",
                  "Annotate",
                  "Composite",
                  "Print",
                  "Write",
                  "Restore",
                  "Load",
                  "Next",
                  "Former",
                  "Quit"
                };

              static KeySym
                MenuCommand[]=
                {
                  XK_i,
                  XK_equal,
                  XK_c,
                  XK_bar,
                  XK_r,
                  XK_g,
                  XK_a,
                  XK_x,
                  XK_p,
                  XK_w,
                  XK_o,
                  XK_l,
                  XK_n,
                  XK_f,
                  XK_q
                };

              /*
                Select a command from the pop-up menu.
              */
              command_number=XMenuWidget(display,window,event.xbutton.x_root,
                event.xbutton.y_root,"Commands",MenuSelections,
                sizeof(MenuSelections)/sizeof(MenuSelections[0]),command);
              if (command_number >= 0)
                loaded_image=XImageWindowCommand(display,resource_info,window,
                  MenuCommand[command_number],&displayed_image,state);
              break;
            }
            case Button2:
            {
              /*
                User pressed the image magnify button.
              */
              if (window->magnify.mapped)
                XRaiseWindow(display,window->magnify.id);
              else
                {
                  /*
                    Make magnify image.
                  */
                  status=XMakeImage(display,resource_info,&window->magnify,
                    (Image *) NULL,window->magnify.width,
                    window->magnify.height);
                  status|=XMakePixmap(display,resource_info,&window->magnify);
                  if (status == False)
                    Error("Unable to create magnify image",(char *) NULL);
                  XMapRaised(display,window->magnify.id);
                }
              XMagnifyImageWindow(display,resource_info,window,&event);
              break;
            }
            case Button3:
            {
              if (displayed_image->montage != (char *) NULL)
                {
                  /*
                    User picked an image tile to display.
                  */
                  loaded_image=XTileImageWindow(display,resource_info,window,
                    displayed_image,&event);
                  if (loaded_image != (Image *) NULL)
                    *state|=NextImageState | ExitState;
                  montage_info.x=window->image.x;
                  montage_info.y=window->image.y;
                  break;
                }
              break;
            }
            default:
              break;
          }
        if (event.xbutton.window == window->pan.id)
          {
            XPanImageWindow(display,window,&event);
            break;
          }
        break;
      }
      case ClientMessage:
      {
        /*
          If client window delete message, exit.
        */
        if (event.xclient.message_type != protocols_property)
          break;
        if (*event.xclient.data.l != delete_property)
          break;
        XWithdrawWindow(display,event.xclient.window,visual_info->screen);
        if (event.xclient.window == window->image.id)
          {
            *state|=ExitState;
            break;
          }
        if (event.xclient.window == window->pan.id)
          {
            /*
              Restore original image size when pan window is deleted.
            */
            window_changes.width=window->image.ximage->width;
            window_changes.height=window->image.ximage->height;
            XReconfigureWMWindow(display,window->image.id,window->image.screen,
              CWWidth | CWHeight,&window_changes);
          }
        break;
      }
      case ConfigureNotify:
      {
        if (resource_info->debug)
          (void) fprintf(stderr,"Configure Notify: 0x%lx %dx%d+%d+%d\n",
            event.xconfigure.window,event.xconfigure.width,
            event.xconfigure.height,event.xconfigure.x,event.xconfigure.y);
        if (event.xconfigure.window == window->image.id)
          {
            /*
              Image window has a new configuration.
            */
            if ((event.xconfigure.width != window->image.width) ||
                (event.xconfigure.height != window->image.height))
              {
                window->image.width=event.xconfigure.width;
                window->image.height=event.xconfigure.height;
                window->image.x=0;
                window->image.y=0;
                if (displayed_image->montage != (char *) NULL)
                  {
                    window->image.x=montage_info.x;
                    window->image.y=montage_info.y;
                  }
                if (window->image.mapped && window->image.stasis)
                  {
                    window->image.window_changes.width=event.xconfigure.width;
                    window->image.window_changes.height=event.xconfigure.height;
                    *state|=UpdateConfigurationState;
                  }
                else
                  if ((event.xconfigure.width < window->image.ximage->width) ||
                      (event.xconfigure.height < window->image.ximage->height))
                    {
                      XMapRaised(display,window->pan.id);
                      XDrawPanRectangle(display,window);
                    }
                  else
                    if (window->pan.mapped)
                      XWithdrawWindow(display,window->pan.id,
                        window->pan.screen);
              }
            break;
          }
        if (event.xconfigure.window == window->magnify.id)
          {
            unsigned int
              magnify;

            /*
              Magnify window has a new configuration.
            */
            window->magnify.width=event.xconfigure.width;
            window->magnify.height=event.xconfigure.height;
            if (!window->magnify.mapped)
              break;
            magnify=1;
            while (magnify <= event.xconfigure.width)
              magnify<<=1;
            while (magnify <= event.xconfigure.height)
              magnify<<=1;
            magnify>>=1;
            if ((magnify != event.xconfigure.width) ||
                (magnify != event.xconfigure.height))
              {
                window_changes.width=magnify;
                window_changes.height=magnify;
                XReconfigureWMWindow(display,window->magnify.id,
                  window->magnify.screen,CWWidth | CWHeight,&window_changes);
                break;
              }
            status=XMakeImage(display,resource_info,&window->magnify,
              (Image *) NULL,window->magnify.width,window->magnify.height);
            status|=XMakePixmap(display,resource_info,&window->magnify);
            if (status == False)
              Error("Unable to create magnify image",(char *) NULL);
            XMakeMagnifyImage(display,resource_info,window);
            break;
          }
        if (event.xconfigure.window == window->pan.id)
          {
            /*
              Icon window has a new configuration.
            */
            window->pan.width=event.xconfigure.width;
            window->pan.height=event.xconfigure.height;
            break;
          }
        if (event.xconfigure.window == window->icon.id)
          {
            /*
              Icon window has a new configuration.
            */
            window->icon.width=event.xconfigure.width;
            window->icon.height=event.xconfigure.height;
            break;
          }
        break;
      }
      case EnterNotify:
      {
        /*
          Selectively install colormap.
        */
        if (map_info->colormap != XDefaultColormap(display,visual_info->screen))
          if (event.xcrossing.mode != NotifyUngrab)
            XInductColormap(display,map_info->colormap);
        break;
      }
      case Expose:
      {
        if (resource_info->debug)
          (void) fprintf(stderr,"Expose: 0x%lx %dx%d+%d+%d\n",
            event.xexpose.window,event.xexpose.width,event.xexpose.height,
            event.xexpose.x,event.xexpose.y);
        /*
          Refresh windows that are now exposed.
        */
        if (event.xexpose.window == window->image.id)
          if (window->image.mapped)
            {
              XRefreshWindow(display,&window->image,&event);
              timer=time((time_t *) NULL)+resource_info->delay;
              break;
            }
        if (event.xexpose.window == window->magnify.id)
          if (event.xexpose.count == 0)
            if (window->magnify.mapped)
              {
                XMakeMagnifyImage(display,resource_info,window);
                break;
              }
        break;
      }
      case FocusOut:
      {
        /*
          Set input focus for backdrop window.
        */
        if (resource_info->debug)
          (void) fprintf(stderr,"FocusOut Notify: 0x%lx\n",event.xfocus.window);
        if (event.xfocus.window == window->image.id)
          XSetInputFocus(display,window->image.id,RevertToNone,CurrentTime);
        break;
      }
      case KeyPress:
      {
        static int
          length;

        /*
          Respond to a user key press.
        */
        length=XLookupString((XKeyEvent *) &event.xkey,command,sizeof(command),
          &key_symbol,(XComposeStatus *) NULL);
        *(command+length)='\0';
        if (resource_info->debug)
          (void) fprintf(stderr,"Key press: 0x%lx (%s)\n",key_symbol,command);
        if (event.xkey.state & ControlMask)
          *state|=ControlState;
        if (event.xkey.window == window->image.id)
          loaded_image=XImageWindowCommand(display,resource_info,window,
            key_symbol,&displayed_image,state);
        if (event.xkey.window == window->magnify.id)
          XMagnifyWindowCommand(display,resource_info,window,key_symbol);
        break;
      }
      case KeyRelease:
      {
        /*
          Respond to a user key release.
        */
        (void) XLookupString((XKeyEvent *) &event.xkey,command,sizeof(command),
          &key_symbol,(XComposeStatus *) NULL);
        if (resource_info->debug)
          (void) fprintf(stderr,"Key release: 0x%lx (%c)\n",key_symbol,
            *command);
        *state&=(~ControlState);
        break;
      }
      case LeaveNotify:
      {
        /*
          Selectively uninstall colormap.
        */
        if (map_info->colormap != XDefaultColormap(display,visual_info->screen))
          if (event.xcrossing.mode != NotifyUngrab)
            XUninductColormap(display,map_info->colormap);
        break;
      }
      case MapNotify:
      {
        if (resource_info->debug)
          (void) fprintf(stderr,"Map Notify: 0x%lx\n",event.xmap.window);
        if (event.xmap.window == window->image.id)
          {
            if (window->backdrop.id != (Window) NULL)
              {
                /*
                  Install colormap and set input focus for backdrop window.
                */
                XInstallColormap(display,map_info->colormap);
                XSetInputFocus(display,window->image.id,RevertToNone,
                  CurrentTime);
              }
            window->image.mapped=True;
            break;
          }
        if (event.xmap.window == window->magnify.id)
          {
            window->magnify.mapped=True;
            break;
          }
        if (event.xmap.window == window->pan.id)
          {
            XMakePanImage(display,resource_info,window,displayed_image);
            window->pan.mapped=True;
            break;
          }
        if (event.xmap.window == window->info.id)
          {
            window->info.mapped=True;
            break;
          }
        if (event.xmap.window == window->icon.id)
          {
            /*
              Create an icon image.
            */
            XMakeStandardColormap(display,icon_visual,&icon_resources,
              displayed_image,icon_map,&icon_pixel);
            (void) XMakeImage(display,&icon_resources,&window->icon,
              displayed_image,window->icon.width,window->icon.height);
            (void) XMakePixmap(display,&icon_resources,&window->icon);
            XSetWindowBackgroundPixmap(display,window->icon.id,
              window->icon.pixmap);
            XClearWindow(display,window->icon.id);
            window->icon.mapped=True;
            break;
          }
        break;
      }
      case MappingNotify:
      {
        XRefreshKeyboardMapping(&event.xmapping);
        break;
      }
      case NoExpose:
        break;
      case ReparentNotify:
      {
        if (resource_info->debug)
          (void) fprintf(stderr,"Reparent Notify: 0x%lx=>0x%lx\n",
            event.xreparent.parent,event.xreparent.window);
        break;
      }
      case UnmapNotify:
      {
        if (resource_info->debug)
          (void) fprintf(stderr,"Unmap Notify: 0x%lx\n",event.xunmap.window);
        if (event.xunmap.window == window->image.id)
          {
            window->image.mapped=False;
            break;
          }
        if (event.xunmap.window == window->magnify.id)
          {
            window->magnify.mapped=False;
            break;
          }
        if (event.xunmap.window == window->pan.id)
          {
            window->pan.mapped=False;
            break;
          }
        if (event.xunmap.window == window->info.id)
          {
            window->info.mapped=False;
            break;
          }
        if (event.xunmap.window == window->icon.id)
          {
            if (map_info->colormap == icon_map->colormap)
              {
                /*
                  Update image colormap.
                */
                window->image.window_changes.width=window->image.width;
                window->image.window_changes.height=window->image.height;
                *state|=UpdateColormapState;
                *state|=UpdateConfigurationState;
              }
            XFreeStandardColormap(display,icon_visual,icon_map,&icon_pixel);
            window->icon.mapped=False;
            break;
          }
        break;
      }
      default:
      {
        if (resource_info->debug)
          (void) fprintf(stderr,"Event type: %d\n",event.type);
        break;
      }
    }
    if (*state & UpdateColormapState)
      {
        /*
          Update window colormap and graphic context.
        */
        XDefineCursor(display,window->image.id,window->image.busy_cursor);
        XFlush(display);
        XMakeStandardColormap(display,visual_info,resource_info,displayed_image,
          map_info,&pixel_info);
        for (i=0; i < number_windows; i++)
        {
          if (magick_windows[i]->id == window->icon.id)
            continue;
          XSetWindowColormap(display,magick_windows[i]->id,map_info->colormap);
          XSetBackground(display,magick_windows[i]->annotate_context,
            pixel_info.background_color.pixel);
          XSetForeground(display,magick_windows[i]->annotate_context,
            pixel_info.foreground_color.pixel);
          XSetBackground(display,magick_windows[i]->widget_context,
            pixel_info.background_color.pixel);
          XSetForeground(display,magick_windows[i]->widget_context,
            pixel_info.foreground_color.pixel);
          XSetBackground(display,magick_windows[i]->highlight_context,
            pixel_info.foreground_color.pixel);
          XSetForeground(display,magick_windows[i]->highlight_context,
            pixel_info.background_color.pixel);
        }
        if (window->backdrop.id != (Window) NULL)
          XInstallColormap(display,map_info->colormap);
        XDefineCursor(display,window->image.id,window->image.cursor);
        *state&=(~UpdateColormapState);
      }
    if (*state & UpdateConfigurationState)
      {
        /*
          Update image window configuration.
        */
        status=XConfigureImageWindow(display,resource_info,window,
          displayed_image);
        if (status == False)
          XNoticeWidget(display,resource_info,window,
            "Unable to configure X image",window->image.name);
        *state&=(~UpdateConfigurationState);
      }
  }
  while (!(*state & ExitState));
  if ((*state & FormerImageState) || (*state & NextImageState))
    *state&=(~ExitState);
  former_ximage=(*window->image.ximage);
  /*
    Alert user we are busy.
  */
  XDefineCursor(display,window->image.id,window->image.busy_cursor);
  XFlush(display);
  if (resource_info->write_filename != (char *) NULL)
    {
      /*
        Update image with user transforms.
      */
      if ((window->image.clip_geometry != (char *) NULL) ||
          (displayed_image->columns != window->image.ximage->width) ||
          (displayed_image->rows != window->image.ximage->height))
        {
          char
            image_geometry[MaxTextLength];

          /*
            Clip and/or scale displayed_image.
          */
          (void) sprintf(image_geometry,"%dx%d",window->image.ximage->width,
            window->image.ximage->height);
          TransformImage(&displayed_image,window->image.clip_geometry,
            image_geometry);
        }
      if (resource_info->number_colors != 0)
        {
          /*
            Reduce the number of colors in the image.
          */
          if ((displayed_image->class == DirectClass) ||
              (displayed_image->colors > resource_info->number_colors) ||
              (resource_info->colorspace == GRAYColorspace))
            QuantizeImage(displayed_image,resource_info->number_colors,
              resource_info->tree_depth,resource_info->dither,
              resource_info->colorspace,True);
          SyncImage(displayed_image);
        }
    }
  /*
    Withdraw pan and magnify window.
  */
  if (window->pan.mapped)
    XWithdrawWindow(display,window->pan.id,window->pan.screen);
  if (window->magnify.mapped)
    XWithdrawWindow(display,window->magnify.id,window->magnify.screen);
  XDelay(display,SuspendTime);
  while (XCheckTypedWindowEvent(display,window->image.id,Expose,&event))
    XRefreshWindow(display,&window->image,&event);
  /*
    Free X resources.
  */
  (void) free((char *) window->popup.name);
  (void) free((char *) window->magnify.name);
  (void) free((char *) window->image.name);
  (void) free((char *) window->image.icon_name);
  for (i=0; i < number_windows; i++)
  {
    XFreeCursor(display,magick_windows[i]->cursor);
    XFreeCursor(display,magick_windows[i]->busy_cursor);
  }
  if (window->popup.highlight_stipple != (Pixmap) NULL)
    XFreePixmap(display,window->popup.highlight_stipple);
  if (window->popup.highlight_stipple != (Pixmap) NULL)
    XFreePixmap(display,window->popup.shadow_stipple);
  XFreeGC(display,pixel_info.highlight_context);
  XFreeGC(display,pixel_info.widget_context);
  XFreeGC(display,pixel_info.annotate_context);
  XFreeFont(display,font_info);
  if (*state & ExitState)
    {
      /*
        Destroy X windows.
      */
      for (i=0; i < number_windows; i++)
      {
        if (magick_windows[i]->id != (Window) NULL)
          XDestroyWindow(display,magick_windows[i]->id);
        if (magick_windows[i]->ximage != (XImage *) NULL)
          XDestroyImage(magick_windows[i]->ximage);
        if (magick_windows[i]->pixmap != (Pixmap) NULL)
          XFreePixmap(display,magick_windows[i]->pixmap);
      }
      /*
        Free Standard Colormap.
      */
      if (resource_info->map_type == (char *) NULL)
        XFreeStandardColormap(display,visual_info,map_info,&pixel_info);
      (void) free((void *) window);
      XFree((void *) class_hint);
      XFree((void *) manager_hints);
      XFree((void *) icon_visual);
      XFree((void *) visual_info);
      XFree((void *) icon_map);
      XFree((void *) map_info);
      visual_info=(XVisualInfo *) NULL;
    }
  XFlush(display);
  *image=displayed_image;
  return(loaded_image);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X D r a w P a n R e c t a n g l e                                         %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XDrawPanRectangle draws a rectangle in the pan window.  The pan
%  window displays a scaled image and the rectangle shows which portion of
%  the image is displayed in the image window.
%
%  The format of the XDrawPanRectangle routine is:
%
%    XDrawPanRectangle(display,window)
%
%  A description of each parameter follows:
%
%    o display: Specifies a connection to an X server;  returned from
%      XOpenDisplay.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%
*/
static void XDrawPanRectangle(display,window)
Display
  *display;

XWindows
  *window;
{
  unsigned long
    scale_factor;

  RectangleInfo
    highlight_info;

  /*
    Determine dimensions of the panning rectangle.
  */
  scale_factor=(unsigned long)
    (UpShift(window->pan.width)/window->image.ximage->width);
  highlight_info.x=DownShift(window->image.x*scale_factor);
  highlight_info.width=DownShift(window->image.width*scale_factor);
  scale_factor=(unsigned long)
    (UpShift(window->pan.height)/window->image.ximage->height);
  highlight_info.y=DownShift(window->image.y*scale_factor);
  highlight_info.height=DownShift(window->image.height*scale_factor);
  /*
    Display the panning rectangle.
  */
  XClearWindow(display,window->pan.id);
  XHighlightRegion(display,window->pan.id,window->pan.annotate_context,
    &highlight_info);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X I m a g e W i n d o w C o m m a n d                                     %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XImageWindowCommand makes a transform to the image or image window
%  as specified by a user menu button or keyboard command.
%
%  The format of the XImageWindowCommand routine is:
%
%    loaded_image=XImageWindowCommand(display,resource_info,window,key_symbol,
%      image,state)
%
%  A description of each parameter follows:
%
%    o loaded_image:  Function XImageWindowCommand returns an image when the
%      user chooses 'Load Image' from the command menu.  Otherwise a null
%      image is returned.
%
%    o display: Specifies a connection to an X server; returned from
%      XOpenDisplay.
%
%    o resource_info: Specifies a pointer to a X11 XResourceInfo structure.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o key_symbol: Specifies a command to perform.
%
%    o image: Specifies a pointer to a Image structure;  XImageWindowCommand
%      may transform the image and return a new image pointer.
%
%    o state: Specifies a pointer to an unsigned long;  XImageWindowCommand
%      may return a modified state.
%
%
*/
static Image *XImageWindowCommand(display,resource_info,window,key_symbol,
  image,state)
Display
  *display;

XResourceInfo
  *resource_info;

XWindows
  *window;

KeySym
  key_symbol;

Image
  **image;

unsigned long
  *state;
{
  char
    geometry[MaxTextLength];

  Image
    *loaded_image;

  unsigned int
    status;

  XEvent
    event;

  /*
    Process user command.
  */
  while (XCheckTypedWindowEvent(display,window->image.id,Expose,&event))
    XRefreshWindow(display,&window->image,&event);
  loaded_image=(Image *) NULL;
  window->image.window_changes.width=window->image.ximage->width;
  window->image.window_changes.height=window->image.ximage->height;
  switch (key_symbol)
  {
    case XK_i:
    {
      char
        *text,
        title[MaxTextLength];

      int
        length;

      /*
        Display information about the image in the Text View widget.
      */
      length=11*MaxTextLength;
      if ((*image)->directory != (char *) NULL)
        length+=strlen((*image)->directory);
      if ((*image)->comments != (char *) NULL)
        length+=strlen((*image)->comments);
      text=(char *) malloc(length*sizeof(char));
      if (text == (char *) NULL)
        {
          XNoticeWidget(display,resource_info,window,
            "Unable to display image info","Memory allocation failed");
          break;
        }
      (void) sprintf(title," Image Info: %s",(*image)->filename);
      (void) sprintf(text,"X\n  visual: %s\n",
        XVisualClassName(window->image.class));
      (void) sprintf(text,"%s  depth: %d\n",text,window->image.depth);
      (void) sprintf(text,"%s  geometry: %dx%d\n",text,
        window->image.ximage->width,window->image.ximage->height);
      if (window->image.clip_geometry != (char *) NULL)
        (void) sprintf(text,"%s  clip geometry: %s\n",text,
          window->image.clip_geometry);
      (void) sprintf(text,"%sImage\n  format: %s\n",text,(*image)->magick);
      if ((*image)->class == DirectClass)
        (void) strcat(text,"  class: DirectClass\n");
      else
        (void) strcat(text,"  class: PseudoClass\n");
      (void) sprintf(text,"%s  geometry: %ux%u\n",text,
        (*image)->columns,(*image)->rows);
      if ((*image)->colors != 0)
        (void) sprintf(text,"%s  colors: %u\n",text,(*image)->colors);
      if ((*image)->signature != (char *) NULL)
        (void) sprintf(text,"%s  signature: %s\n",text,(*image)->signature);
      if ((*image)->scene != 0)
        (void) sprintf(text,"%s  scene: %u\n",text,(*image)->scene);
      if ((*image)->montage != (char *) NULL)
        (void) sprintf(text,"%s  montage: %s\n",text,(*image)->montage);
      if ((*image)->directory != (char *) NULL)
        (void) sprintf(text,"%s  directory:\n\n%s\n",text,(*image)->directory);
      if ((*image)->label != (char *) NULL)
        (void) sprintf(text,"%s  label: %s\n",text,(*image)->label);
      if ((*image)->comments != (char *) NULL)
        (void) sprintf(text,"%s  comments:\n\n%s",text,(*image)->comments);
      XTextViewWidget(display,resource_info,window,title,text);
      (void) free((char *) text);
      break;
    }
    case XK_less:
    {
      /*
        Half image size.
      */
      window->image.window_changes.width=window->image.ximage->width >> 1;
      window->image.window_changes.height=window->image.ximage->height >> 1;
      *state|=UpdateConfigurationState;
      break;
    }
    case XK_equal:
    {
      unsigned int
        height,
        width;

      /*
        Resize image.
      */
      width=window->image.ximage->width;
      height=window->image.ximage->height;
      (void) sprintf(geometry,"%ux%u",width,height);
      XDialogWidget(display,resource_info,window,"Resize",
        "Enter resize geometry (e.g. 640x480, 200%):",geometry);
      if (*geometry == '\0')
        break;
      ParseImageGeometry(geometry,&width,&height);
      window->image.window_changes.width=width;
      window->image.window_changes.height=height;
      *state|=UpdateConfigurationState;
      break;
    }
    case XK_greater:
    {
      /*
        Double the image size.
      */
      window->image.window_changes.width=window->image.ximage->width << 1;
      window->image.window_changes.height=window->image.ximage->height << 1;
      *state|=UpdateConfigurationState;
      break;
    }
    case XK_c:
    {
      /*
        Clip image.
      */
      XClipImageWindow(display,resource_info,window,*image);
      *state|=UpdateConfigurationState;
      break;
    }
    case XK_slash:
    {
      /*
        Rotate image 90 degrees clockwise.
      */
      status=XRotateImageWindow(display,window,90.0,image);
      if (status == False)
        {
          XNoticeWidget(display,resource_info,window,"Unable to rotate X image",
            (*image)->filename);
          break;
        }
      if (window->image.stasis)
        break;
      window->image.window_changes.width=window->image.ximage->height;
      window->image.window_changes.height=window->image.ximage->width;
      *state|=UpdateConfigurationState;
      break;
    }
    case XK_bar:
    {
      /*
        Rotate image.
      */
      status=XRotateImageWindow(display,window,0.0,image);
      if (status == False)
        {
          XNoticeWidget(display,resource_info,window,"Unable to rotate X image",
            (*image)->filename);
          break;
        }
      if (window->image.stasis)
        break;
      window->image.window_changes.width=(*image)->columns;
      window->image.window_changes.height=(*image)->rows;
      if (window->image.clip_geometry != (char *) NULL)
        {
          int
            x,
            y;

          unsigned int
            height,
            width;

          (void) XParseGeometry(window->image.clip_geometry,&x,&y,
            &width,&height);
          window->image.window_changes.width=width;
          window->image.window_changes.height=height;
        }
      *state|=UpdateColormapState;
      *state|=UpdateConfigurationState;
      break;
    }
    case XK_backslash:
    {
      /*
        Rotate image 90 degrees counter-clockwise.
      */
      status=XRotateImageWindow(display,window,-90.0,image);
      if (status == False)
        {
          XNoticeWidget(display,resource_info,window,"Unable to rotate X image",
            (*image)->filename);
          break;
        }
      if (window->image.stasis)
        break;
      window->image.window_changes.width=window->image.ximage->height;
      window->image.window_changes.height=window->image.ximage->width;
      *state|=UpdateConfigurationState;
      break;
    }
    case XK_r:
    {
      /*
        Reflect image scanlines.
      */
      status=XReflectImageWindow(display,window,image);
      if (status == False)
        {
          XNoticeWidget(display,resource_info,window,
            "Unable to reflect X image",(*image)->filename);
          break;
        }
      *state|=UpdateConfigurationState;
      break;
    }
    case XK_g:
    {
      static char
        gamma[MaxTextLength] = "1.6";

      /*
        Gamma correct image.
      */
      XDialogWidget(display,resource_info,window,"Gamma","Enter gamma value:",
        gamma);
      if (*gamma == '\0')
        break;
      GammaImage(*image,gamma);
      *state|=UpdateColormapState;
      *state|=UpdateConfigurationState;
      break;
    }
    case XK_a:
    {
      /*
        Annotate the image with text.
      */
      status=XAnnotateImageWindow(display,resource_info,window,*image);
      if (status == False)
        {
          XNoticeWidget(display,resource_info,window,
            "Unable to annotate X image",(*image)->filename);
          break;
        }
      if (window->image.stasis)
        break;
      *state|=UpdateColormapState;
      *state|=UpdateConfigurationState;
      break;
    }
    case XK_x:
    {
      /*
        Composite image.
      */
      status=XCompositeImageWindow(display,resource_info,window,image);
      if (status == False)
        {
          XNoticeWidget(display,resource_info,window,
            "Unable to composite X image",(*image)->filename);
          break;
        }
      if (window->image.stasis)
        break;
      *state|=UpdateColormapState;
      *state|=UpdateConfigurationState;
      break;
    }
    case XK_p:
    {
      /*
        Print image.
      */
      status=XPrintImageWindow(display,resource_info,window,*image);
      if (status == False)
        {
          XNoticeWidget(display,resource_info,window,"Unable to print X image",
            (*image)->filename);
          break;
        }
      break;
    }
    case XK_w:
    {
      /*
        Write image.
      */
      status=XWriteImageWindow(display,resource_info,window,*image);
      if (status == False)
        {
          XNoticeWidget(display,resource_info,window,"Unable to write X image",
            (*image)->filename);
          break;
        }
      break;
    }
    case XK_o:
    {
      /*
        Restore image window to its original size.
      */
      if (window->image.clip_geometry != (char *) NULL)
        {
          (void) free((char *) window->image.clip_geometry);
          window->image.clip_geometry=(char *) NULL;
          window->image.x=0;
          window->image.y=0;
        }
      window->image.window_changes.width=(*image)->columns;
      window->image.window_changes.height=(*image)->rows;
      *state|=UpdateConfigurationState;
      break;
    }
    case XK_l:
    {
      /*
        Load image.
      */
      loaded_image=XLoadImageWindow(display,resource_info,window,*image);
      if (loaded_image != (Image *) NULL)
        *state|=NextImageState | ExitState;
      break;
    }
    case XK_n:
    case XK_Next:
    case XK_space:
    {
      /*
        Display next image.
      */
      *state|=NextImageState | ExitState;
      break;
    }
    case XK_f:
    case XK_Prior:
    case XK_BackSpace:
    {
      /*
        Display former image.
      */
      *state|=FormerImageState | ExitState;
      break;
    }
    case XK_q:
    {
      /*
        Exit program.
      */
      *state|=ExitState;  /* exit program */
      break;
    }
    case XK_Help:
    {
      Usage(False);
      break;
    }
    case XK_Home:
    {
      XTranslateImageWindow(display,window,*image,key_symbol);
      break;
    }
    case XK_Up:
    case XK_Down:
    case XK_Left:
    case XK_Right:
    {
      if (*state & ControlState)
        {
          RectangleInfo
            clip_info;

          /*
            Trim one pixel from edge of image.
          */
          clip_info.x=0;
          clip_info.y=0;
          clip_info.width=window->image.ximage->width;
          clip_info.height=window->image.ximage->height;
          if (key_symbol == XK_Up)
            clip_info.height--;
          if (key_symbol == XK_Down)
            {
              clip_info.y++;
              clip_info.height--;
            }
          if (key_symbol == XK_Right)
            clip_info.width--;
          if (key_symbol == XK_Left)
            {
              clip_info.x++;
              clip_info.width--;
            }
          XSetClipGeometry(display,window,&clip_info,*image);
          window->image.window_changes.width=clip_info.width;
          window->image.window_changes.height=clip_info.height;
          /*
            Prevent image flashing-- do not repaint window background.
          */
          XSetWindowBackgroundPixmap(display,window->image.id,None);
          *state|=UpdateConfigurationState;
          break;
        }
      XTranslateImageWindow(display,window,*image,key_symbol);
      break;
    }
    case XK_Return:
    case XK_KP_Enter:
      break;
    default:
    {
      if (!IsModifierKey(key_symbol))
        XBell(display,0);
      break;
    }
  }
  return(loaded_image);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X L o a d I m a g e W i n d o w                                           %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XLoadImageWindow loads an image from a file.
%
%  The format of the XLoadImageWindow routine is:
%
%    loaded_image=XLoadImageWindow(display,resource_info,window,image)
%
%  A description of each parameter follows:
%
%    o status: Function XLoadImageWindow returns an image if can be loaded
%      successfully.  Otherwise a null image is returned.
%
%    o display: Specifies a connection to an X server; returned from
%      XOpenDisplay.
%
%    o resource_info: Specifies a pointer to a X11 XResourceInfo structure.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o image: Specifies a pointer to a Image structure;  returned from
%      ReadImage.
%
%
*/
static Image *XLoadImageWindow(display,resource_info,window,image)
Display
  *display;

XResourceInfo
  *resource_info;

XWindows
  *window;

Image
  *image;
{
  char
    filename[MaxTextLength],
    text[MaxTextLength];

  Image
    *loaded_image;

  /*
    Request file name from user.
  */
  (void) strcpy(filename,image->filename);
  XFileBrowserWidget(display,resource_info,window,"Load",filename);
  if (*filename == '\0')
    return((Image *) NULL);
  /*
    Map info window.
  */
  (void) strcpy(text," Loading image... ");
  XSetWindowExtents(display,&window->info,text);
  XMapWindow(display,window->info.id);
  XDisplayInfoString(display,&window->info,text);
  /*
    Load the image.
  */
  XDefineCursor(display,window->image.id,window->image.busy_cursor);
  XFlush(display);
  (void) strcpy(resource_info->image_info->filename,filename);
  loaded_image=ReadImage(resource_info->image_info);
  XDefineCursor(display,window->image.id,window->image.cursor);
  XWithdrawWindow(display,window->info.id,window->info.screen);
  if (loaded_image == (Image *) NULL)
    XNoticeWidget(display,resource_info,window,"Unable to load image",filename);
  return(loaded_image);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X M a g n i f y I m a g e W i n d o w                                     %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XMagnifyImageWindow magnifies portions of the image as indicated
%  by the pointer.  The magnified portion is displayed in a separate window.
%
%  The format of the XMagnifyImageWindow routine is:
%
%    XMagnifyImageWindow(display,resource_info,window,event)
%
%  A description of each parameter follows:
%
%    o display: Specifies a connection to an X server;  returned from
%      XOpenDisplay.
%
%    o resource_info: Specifies a pointer to a X11 XResourceInfo structure.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o event: Specifies a pointer to a XEvent structure.  If it is NULL,
%      the entire image is refreshed.
%
%
*/
static void XMagnifyImageWindow(display,resource_info,window,event)
Display
  *display;

XResourceInfo
  *resource_info;

XWindows
  *window;

XEvent
  *event;
{
  char
    text[MaxTextLength];

  register int
    x,
    y;

  unsigned long
    state;

  /*
    Map info window.
  */
  state=DefaultState;
  (void) sprintf(text," +%u+%u ",window->image.width,window->image.height);
  XSetWindowExtents(display,&window->info,text);
  XMapWindow(display,window->info.id);
  window->info.mapped=True;
  /*
    Update magnified image until the mouse button is released.
  */
  XDefineCursor(display,window->image.id,window->magnify.cursor);
  x=event->xbutton.x;
  y=event->xbutton.y;
  window->magnify.x=x+1;
  window->magnify.y=y+1;
  do
  {
    /*
      Check boundary conditions.
    */
    if (x < 0)
      x=0;
    else
      if (x >= window->image.width)
        x=window->image.width-1;
    if (y < 0)
      y=0;
    else
     if (y >= window->image.height)
       y=window->image.height-1;
    if ((window->magnify.x != (window->image.x+x)) ||
        (window->magnify.y != (window->image.y+y)))
      {
        window->magnify.x=window->image.x+x;
        window->magnify.y=window->image.y+y;
        /*
          Map and unmap info window as text cursor crosses its boundaries.
        */
        if (window->info.mapped)
          {
            if ((x < (window->info.x+window->info.width)) &&
                (y < (window->info.y+window->info.height)))
              {
                XWithdrawWindow(display,window->info.id,window->info.screen);
                window->info.mapped=False;
              }
          }
        else
          if ((x > (window->info.x+window->info.width)) ||
              (y > (window->info.y+window->info.height)))
            {
              XMapWindow(display,window->info.id);
              window->info.mapped=True;
            }
        if (window->info.mapped)
          {
            /*
              Display pointer position.
            */
            (void) sprintf(text," %+d%+d ",window->magnify.x,window->magnify.y);
            XDisplayInfoString(display,&window->info,text);
          }
        /*
          Display magnified image.
        */
        XMakeMagnifyImage(display,resource_info,window);
      }
    /*
      Wait for next event.
    */
    XMaskEvent(display,ButtonPressMask | ButtonMotionMask | ButtonReleaseMask |
      ExposureMask,event);
    switch (event->type)
    {
      case ButtonPress:
        break;
      case ButtonRelease:
      {
        /*
          User has finished magnifying image.
        */
        if (event->xbutton.button != Button2)
          break;
        x=event->xbutton.x;
        y=event->xbutton.y;
        state|=ExitState;
        break;
      }
      case Expose:
      {
        /*
          Refresh image window.
        */
        XRefreshWindow(display,&window->image,event);
        break;
      }
      case MotionNotify:
      {
        /*
          Discard pending button motion events.
        */
        while (XCheckMaskEvent(display,ButtonMotionMask,event));
        x=event->xmotion.x;
        y=event->xmotion.y;
        break;
      }
      default:
        break;
    }
  } while (!(state & ExitState));
  /*
    Check boundary conditions.
  */
  if (x < 0)
    x=0;
  else
    if (x >= window->image.width)
      x=window->image.width-1;
  window->magnify.x=window->image.x+x;
  if (y < 0)
    y=0;
  else
   if (y >= window->image.height)
     y=window->image.height-1;
  window->magnify.y=window->image.y+y;
  /*
    Display magnified image.
  */
  XMakeMagnifyImage(display,resource_info,window);
  /*
    Restore cursor.
  */
  XDefineCursor(display,window->image.id,window->image.cursor);
  if (window->info.mapped)
    XWithdrawWindow(display,window->info.id,window->info.screen);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X M a g n i f y W i n d o w C o m m a n d                                 %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XMagnifyWindowCommand moves the image within the magnify window by
%  one pixel as specified by the key symbol.
%
%  The format of the XMagnifyWindowCommand routine is:
%
%    XMagnifyWindowCommand(display,resource_info,window,key_symbol)
%
%  A description of each parameter follows:
%
%    o display: Specifies a connection to an X server; returned from
%      XOpenDisplay.
%
%    o resource_info: Specifies a pointer to a X11 XResourceInfo structure.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o key_symbol: Specifies a KeySym which indicates which side of the image
%      to trim.
%
%
*/
static void XMagnifyWindowCommand(display,resource_info,window,key_symbol)
Display
  *display;

XResourceInfo
  *resource_info;

XWindows
  *window;

KeySym
  key_symbol;
{
  /*
    User specified a magnify factor or position.
  */
  switch (key_symbol)
  {
    case XK_Home:
    {
      window->magnify.x=window->image.width >> 1;
      window->magnify.y=window->image.height >> 1;
      break;
    }
    case XK_Left:
    {
      if (window->magnify.x > 0)
        window->magnify.x--;
      break;
    }
    case XK_Up:
    {
      if (window->magnify.y > 0)
        window->magnify.y--;
      break;
    }
    case XK_Right:
    {
      if (window->magnify.x < (window->image.width-1))
        window->magnify.x++;
      break;
    }
    case XK_Down:
    {
      if (window->magnify.y < (window->image.height-1))
        window->magnify.y++;
      break;
    }
    case XK_0:
    case XK_1:
    case XK_2:
    case XK_3:
    case XK_4:
    case XK_5:
    case XK_6:
    case XK_7:
    case XK_8:
    case XK_9:
    {
      resource_info->magnify=key_symbol-XK_0;
      break;
    }
    default:
      break;
  }
  XMakeMagnifyImage(display,resource_info,window);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X M a k e P a n I m a g e                                                 %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XMakePanImage creates a thumbnail of the image and displays it in
%  the pan icon window.
%
%  The format of the XMakePanImage routine is:
%
%      XMakePanImage(display,resource_info,window,image)
%
%  A description of each parameter follows:
%
%    o display: Specifies a connection to an X server;  returned from
%      XOpenDisplay.
%
%    o resource_info: Specifies a pointer to a X11 XResourceInfo structure.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o image: Specifies a pointer to a Image structure;  returned from
%      ReadImage.
%
%
*/
static void XMakePanImage(display,resource_info,window,image)
Display
  *display;

XResourceInfo
  *resource_info;

XWindows
  *window;

Image
  *image;
{
  unsigned int
    status;

  /*
    Create and display image for panning icon.
  */
  XDefineCursor(display,window->image.id,window->image.busy_cursor);
  XFlush(display);
  window->pan.x=window->image.x;
  window->pan.y=window->image.y;
  status=XMakeImage(display,resource_info,&window->pan,image,window->pan.width,
    window->pan.height);
  status|=XMakePixmap(display,resource_info,&window->pan);
  if (status == False)
    Error("Unable to create pan icon image",(char *) NULL);
  XSetWindowBackgroundPixmap(display,window->pan.id,window->pan.pixmap);
  XClearWindow(display,window->pan.id);
  XDrawPanRectangle(display,window);
  XDefineCursor(display,window->image.id,window->image.cursor);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X P a n I m a g e W i n d o w                                             %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XPanImageWindow pans the image until the mouse button is released.
%
%  The format of the XPanImageWindow routine is:
%
%    XPanImageWindow(display,window,event)
%
%  A description of each parameter follows:
%
%    o display: Specifies a connection to an X server;  returned from
%      XOpenDisplay.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o event: Specifies a pointer to a XEvent structure.  If it is NULL,
%      the entire image is refreshed.
%
*/
static void XPanImageWindow(display,window,event)
Display
  *display;

XWindows
  *window;

XEvent
  *event;
{
  char
    text[MaxTextLength];

  Cursor
    cursor;

  unsigned long
    state;

  unsigned long
    x_factor,
    y_factor;

  XRectangle
    pan_info;

  /*
    Map info window.
  */
  state=DefaultState;
  (void) sprintf(text," %ux%u%+d%+d  ",window->image.width,window->image.height,
    window->image.ximage->width,window->image.ximage->height);
  XSetWindowExtents(display,&window->info,text);
  XMapWindow(display,window->info.id);
  window->info.mapped=True;
  /*
    Define cursor.
  */
  if ((window->image.ximage->width > window->image.width) &&
      (window->image.ximage->height > window->image.height))
    cursor=XCreateFontCursor(display,XC_fleur);
  else
    if (window->image.ximage->width > window->image.width)
      cursor=XCreateFontCursor(display,XC_sb_h_double_arrow);
    else
      if (window->image.ximage->height > window->image.height)
        cursor=XCreateFontCursor(display,XC_sb_v_double_arrow);
      else
        cursor=XCreateFontCursor(display,XC_arrow);
  XDefineCursor(display,window->pan.id,cursor);
  /*
    Pan image as pointer moves until the mouse button is released.
  */
  x_factor=(unsigned long)
    UpShift(window->image.ximage->width)/window->pan.width;
  y_factor=(unsigned long)
    UpShift(window->image.ximage->height)/window->pan.height;
  pan_info.x=event->xbutton.x;
  pan_info.y=event->xbutton.y;
  pan_info.width=(unsigned short) (UpShift(window->image.width)/x_factor);
  pan_info.height=(unsigned short) (UpShift(window->image.height)/y_factor);
  window->image.x=pan_info.x+1;
  window->image.y=pan_info.y+1;
  do
  {
    /*
      Check boundary conditions.
    */
    pan_info.x=DownShift((pan_info.x-(pan_info.width >> 1))*x_factor);
    if (pan_info.x < 0)
      pan_info.x=0;
    else
      if ((pan_info.x+window->image.width) > window->image.ximage->width)
        pan_info.x=window->image.ximage->width-window->image.width;
    pan_info.y=DownShift((pan_info.y-(pan_info.height >> 1))*y_factor);
    if (pan_info.y < 0)
      pan_info.y=0;
    else
      if ((pan_info.y+window->image.height) > window->image.ximage->height)
        pan_info.y=window->image.ximage->height-window->image.height;
    if ((window->image.x != pan_info.x) || (window->image.y != pan_info.y))
      {
        /*
          Display image pan offset.
        */
        window->image.x=pan_info.x;
        window->image.y=pan_info.y;
        (void) sprintf(text," %ux%u%+d%+d ",window->image.width,
          window->image.height,window->image.x,window->image.y);
        XDisplayInfoString(display,&window->info,text);
        /*
          Refresh image window.
        */
        XDrawPanRectangle(display,window);
        XRefreshWindow(display,&window->image,(XEvent *) NULL);
      }
    /*
      Wait for next event.
    */
    XMaskEvent(display,ButtonPressMask | ButtonMotionMask | ButtonReleaseMask,
      event);
    switch (event->type)
    {
      case ButtonRelease:
      {
        /*
          User has finished panning the image.
        */
        state|=ExitState;
        pan_info.x=event->xbutton.x;
        pan_info.y=event->xbutton.y;
        break;
      }
      case MotionNotify:
      {
        /*
          Discard pending button motion events.
        */
        while (XCheckMaskEvent(display,ButtonMotionMask,event));
        pan_info.x=event->xmotion.x;
        pan_info.y=event->xmotion.y;
      }
      default:
        break;
    }
  } while (!(state & ExitState));
  /*
    Check boundary conditions.
  */
  pan_info.x=DownShift((pan_info.x-(pan_info.width >> 1))*x_factor);
  if (pan_info.x < 0)
    pan_info.x=0;
  else
    if ((pan_info.x+window->image.width) > window->image.ximage->width)
      pan_info.x=window->image.ximage->width-window->image.width;
  pan_info.y=DownShift((pan_info.y-(pan_info.height >> 1))*y_factor);
  if (pan_info.y < 0)
    pan_info.y=0;
  else
    if ((pan_info.y+window->image.height) > window->image.ximage->height)
      pan_info.y=window->image.ximage->height-window->image.height;
  if ((window->image.x != pan_info.x) || (window->image.y != pan_info.y))
    {
      /*
        Refresh image window.
      */
      window->image.x=pan_info.x;
      window->image.y=pan_info.y;
      XDrawPanRectangle(display,window);
      XRefreshWindow(display,&window->image,(XEvent *) NULL);
    }
  /*
    Restore cursor.
  */
  XDefineCursor(display,window->pan.id,window->pan.cursor);
  XFreeCursor(display,cursor);
  XWithdrawWindow(display,window->info.id,window->info.screen);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X P r i n t I m a g e W i n d o w                                         %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XPrintImageWindow prints an image to a Postscript printer.
%
%  The format of the XPrintImageWindow routine is:
%
%    status=XPrintImageWindow(display,resource_info,window,image)
%
%  A description of each parameter follows:
%
%    o status: Function XPrintImageWindow return True if the image is
%      printed.  False is returned is there is a memory shortage or if the
%      image fails to print.
%
%    o display: Specifies a connection to an X server; returned from
%      XOpenDisplay.
%
%    o resource_info: Specifies a pointer to a X11 XResourceInfo structure.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o image: Specifies a pointer to a Image structure;  returned from
%      ReadImage.
%
%
*/
static unsigned int XPrintImageWindow(display,resource_info,window,image)
Display
  *display;

XResourceInfo
  *resource_info;

XWindows
  *window;

Image
  *image;
{
  char
    command[MaxTextLength],
    text[MaxTextLength];

  Image
    *output_image;

  int
    status;

  /*
    Request file name from user.
  */
  (void) strcpy(command,resource_info->print_command);
  XDialogWidget(display,resource_info,window,"Print","Print command:",command);
  if (*command == '\0')
    return(True);
  /*
    Alert user we are busy.
  */
  XDefineCursor(display,window->image.id,window->image.busy_cursor);
  (void) strcpy(text," Printing image... ");
  XSetWindowExtents(display,&window->info,text);
  XMapWindow(display,window->info.id);
  XDisplayInfoString(display,&window->info,text);
  XFlush(display);
  /*
    Copy image before applying image transforms.
  */
  image->orphan=True;
  output_image=CopyImage(image,image->columns,image->rows,True);
  image->orphan=False;
  if (output_image == (Image *) NULL)
    {
      XDefineCursor(display,window->image.id,window->image.cursor);
      XWithdrawWindow(display,window->info.id,window->info.screen);
      return(False);
    }
  if ((window->image.clip_geometry != (char *) NULL) ||
     (output_image->columns != window->image.ximage->width) ||
     (output_image->rows != window->image.ximage->height))
    {
      char
        image_geometry[MaxTextLength];

      /*
        Clip and/or scale image.
      */
      (void) sprintf(image_geometry,"%dx%d",window->image.ximage->width,
        window->image.ximage->height);
      TransformImage(&output_image,window->image.clip_geometry,image_geometry);
    }
  if (resource_info->number_colors != 0)
    {
      /*
        Reduce the number of colors in the image.
      */
      if ((output_image->class == DirectClass) ||
          (output_image->colors > resource_info->number_colors) ||
          (resource_info->colorspace == GRAYColorspace))
        QuantizeImage(output_image,resource_info->number_colors,
          resource_info->tree_depth,resource_info->dither,
          resource_info->colorspace,True);
      SyncImage(output_image);
    }
  (void) sprintf(output_image->filename,"ps:|%s",command);
  status=WriteImage(resource_info->image_info,output_image);
  DestroyImage(output_image);
  XDefineCursor(display,window->image.id,window->image.cursor);
  XWithdrawWindow(display,window->info.id,window->info.screen);
  return(status);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X R e f l e c t I m a g e W i n d o w                                     %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XReflectImageWindow reflects the scanlines of an image.
%
%  The format of the XReflectImageWindow routine is:
%
%    status=XReflectImageWindow(display,window,image)
%
%  A description of each parameter follows:
%
%    o status: Function XReflectImageWindow return True if the window scanlines
%      reverse.  False is returned is there is a memory shortage or if the
%      window scanlines fails to reverse.
%
%    o display: Specifies a connection to an X server; returned from
%      XOpenDisplay.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o image: Specifies a pointer to a Image structure;  returned from
%      ReadImage.
%
%
*/
static unsigned int XReflectImageWindow(display,window,image)
Display
  *display;

XWindows
  *window;

Image
  **image;
{
  char
    text[MaxTextLength];

  Image
    *reflected_image;

  if (((*image)->columns*(*image)->rows) > MinInfoSize)
    {
      /*
        Map image window.
      */
      (void) strcpy(text," Reflecting image... ");
      XSetWindowExtents(display,&window->info,text);
      XMapWindow(display,window->info.id);
      XDisplayInfoString(display,&window->info,text);
      window->info.mapped=True;
    }
  /*
    Reflect image scanlines.
  */
  XDefineCursor(display,window->image.id,window->image.busy_cursor);
  XFlush(display);
  reflected_image=ReflectImage(*image);
  XDefineCursor(display,window->image.id,window->image.cursor);
  if (window->info.mapped)
    XWithdrawWindow(display,window->info.id,window->info.screen);
  if (reflected_image == (Image *) NULL)
    return(False);
  DestroyImage(*image);
  *image=reflected_image;
  if (window->image.clip_geometry != (char *) NULL)
    {
      int
        x,
        y;
 
      unsigned int
        height,
        width;
 
      /*
        Reverse clip geometry.
      */
      (void) XParseGeometry(window->image.clip_geometry,&x,&y,&width,&height);
      (void) sprintf(window->image.clip_geometry,"%ux%u%+d%+d",width,height,
        (int) (*image)->columns-(int) width-x,y);
    }
  return(True);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X R o t a t e I m a g e W i n d o w                                       %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XRotateImageWindow rotates the X image.  If the degrees parameter
%  if zero, the rotation angle is computed from the slope of a line drawn by
%  the user.
%
%  The format of the XRotateImageWindow routine is:
%
%    status=XRotateImageWindow(display,window,degrees,image)
%
%  A description of each parameter follows:
%
%    o status: Function XRotateImageWindow return True if the image is
%      rotated.  False is returned is there is a memory shortage or if the
%      image fails to rotate.
%
%    o display: Specifies a connection to an X server; returned from
%      XOpenDisplay.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o degrees: Specifies the number of degrees to rotate the image.
%
%    o image: Specifies a pointer to a Image structure;  returned from
%      ReadImage.
%
%
*/
static unsigned int XRotateImageWindow(display,window,degrees,image)
Display
  *display;

XWindows
  *window;

double
  degrees;

Image
  **image;
{
  char
    text[MaxTextLength];

  Image
    *rotated_image;

  if (degrees == 0.0)
    {
      int
        i,
        x,
        y;

      unsigned int
        distance,
        mask;

      unsigned long
        state;

      Window
        xwindow;

      XEvent
        event;

      XSegment
        rotate_info;

      /*
        Wait for first button press.
      */
      XSetFunction(display,window->image.highlight_context,GXinvert);
      XSelectInput(display,window->image.id,
        window->image.attributes.event_mask | PointerMotionMask);
      XQueryPointer(display,window->image.id,&xwindow,&xwindow,&i,&i,&x,&y,
        &mask);
      rotate_info.x1=x;
      rotate_info.y1=y;
      rotate_info.x2=rotate_info.x1+32;
      rotate_info.y2=rotate_info.y1;
      state=DefaultState;
      do
      {
        XHighlightLine(display,window->image.id,
           window->image.highlight_context,&rotate_info);
        /*
          Wait for next event.
        */
        XMaskEvent(display,ButtonPressMask | ExposureMask | PointerMotionMask,
          &event);
        XHighlightLine(display,window->image.id,
           window->image.highlight_context,&rotate_info);
        switch (event.type)
        {
          case ButtonPress:
          {
            /*
              User has committed to start point of rotation line.
            */
            rotate_info.x1=event.xbutton.x;
            rotate_info.y1=event.xbutton.y;
            state|=ExitState;
            break;
          }
          case Expose:
          {
            /*
              Refresh image window.
            */
            XRefreshWindow(display,&window->image,&event);
            break;
          }
          case MotionNotify:
          {
            /*
              Discard pending pointer motion events.
            */
            while (XCheckMaskEvent(display,PointerMotionMask,&event));
            rotate_info.x1=event.xmotion.x;
            rotate_info.y1=event.xmotion.y;
          }
        }
        rotate_info.x2=rotate_info.x1+32;
        rotate_info.y2=rotate_info.y1;
      } while (!(state & ExitState));
      XSelectInput(display,window->image.id,
        window->image.attributes.event_mask);
      XWithdrawWindow(display,window->info.id,window->info.screen);
      /*
        Draw line as pointer moves until the mouse button is released.
      */
      distance=0;
      (void) sprintf(text," %.2f",-180.0);
      XSetWindowExtents(display,&window->info,text);
      state=DefaultState;
      do
      {
        if (distance > 9)
          {
            /*
              Display info and draw rotation line.
            */
            if (!window->info.mapped)
              {
                XMapWindow(display,window->info.id);
                window->info.mapped=True;
              }
            (void) sprintf(text," %.2f",degrees);
            XDisplayInfoString(display,&window->info,text);
            XHighlightLine(display,window->image.id,
              window->image.highlight_context,&rotate_info);
          }
        else
          if (window->info.mapped)
            {
              /*
                Rotation line is too small;  withdraw info window.
              */
              XWithdrawWindow(display,window->info.id,window->info.screen);
              window->info.mapped=False;
            }
        /*
          Wait for next event.
        */
        XMaskEvent(display,ButtonMotionMask | ButtonPressMask |
          ButtonReleaseMask | ExposureMask,&event);
        if (distance > 9)
          XHighlightLine(display,window->image.id,
            window->image.highlight_context,&rotate_info);
        switch (event.type)
        {
          case ButtonPress:
            break;
          case ButtonRelease:
          {
            /*
              User has committed to rotation line.
            */
            rotate_info.x2=event.xbutton.x;
            rotate_info.y2=event.xbutton.y;
            state|=ExitState;
            break;
          }
          case Expose:
          {
            /*
              Refresh image window.
            */
            XRefreshWindow(display,&window->image,&event);
            break;
          }
          case MotionNotify:
          {
            /*
              Discard pending button motion events.
            */
            while (XCheckMaskEvent(display,ButtonMotionMask,&event));
            rotate_info.x2=event.xmotion.x;
            rotate_info.y2=event.xmotion.y;
          }
          default:
            break;
        }
        /*
          Check boundary conditions.
        */
        if (rotate_info.x2 < 0)
          rotate_info.x2=0;
        else
          if (rotate_info.x2 > window->image.width)
            rotate_info.x2=window->image.width;
        if (rotate_info.y2 < 0)
          rotate_info.y2=0;
        else
          if (rotate_info.y2 > window->image.height)
            rotate_info.y2=window->image.height;
        /*
          Compute rotation angle from the slope of the line.
        */
        degrees=0.0;
        distance=
          ((rotate_info.x2-rotate_info.x1)*(rotate_info.x2-rotate_info.x1))+
          ((rotate_info.y2-rotate_info.y1)*(rotate_info.y2-rotate_info.y1));
        if (distance > 9)
          degrees=RadiansToDegrees(-atan2((double) (rotate_info.y2-
            rotate_info.y1),(double) (rotate_info.x2-rotate_info.x1)));
      } while (!(state & ExitState));
      XSetFunction(display,window->image.highlight_context,GXcopy);
      if (window->info.mapped)
        XWithdrawWindow(display,window->info.id,window->info.screen);
    }
  if (degrees == 0.0)
    return(True);
  if (((*image)->columns*(*image)->rows) > MinInfoSize)
    {
      /*
        Map info window.
      */
      (void) strcpy(text," Rotating image... ");
      XSetWindowExtents(display,&window->info,text);
      XMapWindow(display,window->info.id);
      XDisplayInfoString(display,&window->info,text);
      window->info.mapped=True;
    }
  /*
    Rotate image.
  */
  XDefineCursor(display,window->image.id,window->image.busy_cursor);
  XFlush(display);
  rotated_image=RotateImage(*image,degrees,False);
  XDefineCursor(display,window->image.id,window->image.cursor);
  if (window->info.mapped)
    XWithdrawWindow(display,window->info.id,window->info.screen);
  if (rotated_image == (Image *) NULL)
    return(False);
  DestroyImage(*image);
  *image=rotated_image;
  if (window->image.clip_geometry != (char *) NULL)
    {
      int
        x,
        y;

      unsigned int
        height,
        rotations,
        width;

      /*
        Rotate clip geometry.
      */
      (void) XParseGeometry(window->image.clip_geometry,&x,&y,&width,&height);
      while (degrees < -45.0)
        degrees+=360.0;
      for (rotations=0; degrees > 45.0; rotations++)
        degrees-=90.0;
      switch (rotations % 4)
      {
        default:
        case 0:
          break;
        case 1:
        {
          /*
            Rotate 90 degrees.
          */
          (void) sprintf(window->image.clip_geometry,"%ux%u%+d%+d",height,width,
            (int) (*image)->columns-(int) height-y,x);
          break;
        }
        case 2:
        {
          /*
            Rotate 180 degrees.
          */
          (void) sprintf(window->image.clip_geometry,"%ux%u%+d%+d",width,height,
            (int) width-x,(int) height-y);
          break;
        }
        case 3:
        {
          /*
            Rotate 270 degrees.
          */
          (void) sprintf(window->image.clip_geometry,"%ux%u%+d%+d",height,width,
            y,(int) (*image)->rows-(int) width-x);
          break;
        }
      }
    }
  window->image.stasis=False;
  return(True);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X S e t C l i p G e o m e t r y                                           %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XSetClipGeometry accepts a clipping geometry relative to the
%  image window and translates it to a clipping geometry relative to the
%  image.
%
%  The format of the XSetClipGeometry routine is:
%
%    XSetClipGeometry(display,window,clip_info,image)
%
%  A description of each parameter follows:
%
%    o display: Specifies a connection to an X server; returned from
%      XOpenDisplay.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o clip_info:  A pointer to a RectangleInfo that defines a region of the
%      image window to clip.
%
%    o image: Specifies a pointer to a Image structure.
%
%
*/
static void XSetClipGeometry(display,window,clip_info,image)
Display
  *display;

XWindows
  *window;

RectangleInfo
  *clip_info;

Image
  *image;
{
  char
    text[MaxTextLength];

  int
    x,
    y;

  unsigned int
    height,
    width;

  unsigned long
    scale_factor;

  /*
    Display info on clipping rectangle.
  */
  (void) sprintf(text," %ux%u+%u+%u ",window->image.width,
    window->image.height,window->image.width,window->image.height);
  XSetWindowExtents(display,&window->info,text);
  XMapWindow(display,window->info.id);
  (void) sprintf(text," %ux%u%+d%+d",clip_info->width,clip_info->height,
    clip_info->x,clip_info->y);
  XDisplayInfoString(display,&window->info,text);
  /*
    Clipping geometry is relative to any previous clip geometry.
  */
  x=0;
  y=0;
  width=image->columns;
  height=image->rows;
  if (window->image.clip_geometry != (char *) NULL)
    (void) XParseGeometry(window->image.clip_geometry,&x,&y,&width,&height);
  else
    {
      /*
        Allocate clip geometry string.
      */
      window->image.clip_geometry=(char *) malloc(MaxTextLength*sizeof(char));
      if (window->image.clip_geometry == (char *) NULL)
        Error("Unable to clip X image",window->image.name);
    }
  /*
    Define the clip geometry string from the clipping rectangle.
  */
  scale_factor=UpShift(width)/window->image.ximage->width;
  clip_info->x+=window->image.x;
  if (clip_info->x > 0)
    x+=DownShift(clip_info->x*scale_factor);
  width=DownShift(clip_info->width*scale_factor);
  if (width == 0)
    width=1;
  scale_factor=UpShift(height)/window->image.ximage->height;
  clip_info->y+=window->image.y;
  if (clip_info->y > 0)
    y+=DownShift(clip_info->y*scale_factor);
  height=DownShift(clip_info->height*scale_factor);
  if (height == 0)
    height=1;
  (void) sprintf(window->image.clip_geometry,"%ux%u%+d%+d",width,height,x,y);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X T i l e I m a g e W i n d o w                                           %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XTileImageWindow determines which individual tile of a composite
%  image was choosen with a button press and then displays it.
%
%  The format of the XTileImageWindow routine is:
%
%    tiled_image=XTileImageWindow(display,resource_info,window,image,event)
%
%  A description of each parameter follows:
%
%    o tiled_image:  XTileImageWindow reads the tiled image and returns
%      it.  A null image is returned if an error occurs.
%
%    o display: Specifies a connection to an X server;  returned from
%      XOpenDisplay.
%
%    o resource_info: Specifies a pointer to a X11 XResourceInfo structure.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o image: Specifies a pointer to a Image structure; returned from
%      ReadImage.
%
%    o event: Specifies a pointer to a XEvent structure.  If it is NULL,
%      the entire image is refreshed.
%
%
*/
static Image *XTileImageWindow(display,resource_info,window,image,event)
Display
  *display;

XResourceInfo
  *resource_info;

XWindows
  *window;

Image
  *image;

XEvent
  *event;
{
  char
    filename[MaxTextLength],
    text[MaxTextLength];

  Image
    *tiled_image;

  int
    tile,
    x,
    y;

  register char
    *p,
    *q;

  unsigned int
    height,
    width;

  unsigned long
    scale_factor;

  /*
    Tile image is relative to composite image configuration.
  */
  x=0;
  y=0;
  width=image->columns;
  height=image->rows;
  if (window->image.clip_geometry != (char *) NULL)
    (void) XParseGeometry(window->image.clip_geometry,&x,&y,&width,&height);
  scale_factor=UpShift(width)/window->image.ximage->width;
  event->xbutton.x+=window->image.x;
  event->xbutton.x=DownShift(event->xbutton.x*scale_factor)+x;
  scale_factor=UpShift(height)/window->image.ximage->height;
  event->xbutton.y+=window->image.y;
  event->xbutton.y=DownShift(event->xbutton.y*scale_factor)+y;
  /*
    Determine size and location of individual tiles of the composite.
  */
  x=0;
  y=0;
  width=image->columns;
  height=image->rows;
  (void) XParseGeometry(image->montage,&x,&y,&width,&height);
  tile=((event->xbutton.y-y)/height)*((image->columns-x)/width)+
    (event->xbutton.x-x)/width;
  if (tile < 0)
    {
      /*
        Button press is outside any tile.
      */
      XBell(display,0);
      return((Image *) NULL);
    }
  /*
    Determine file name from the tile directory.
  */
  p=image->directory;
  while ((tile != 0) && (*p != '\0'))
  {
    if (*p == '\n')
      tile--;
    p++;
  }
  if (*p == '\0')
    {
      /*
        Button press is outside any tile.
      */
      XBell(display,0);
      return((Image *) NULL);
    }
  q=p;
  while ((*q != '\n') && (*q != '\0'))
    q++;
  (void) strncpy(filename,p,q-p);
  filename[q-p]='\0';
  /*
    Map info window.
  */
  (void) strcpy(text," Loading image... ");
  XSetWindowExtents(display,&window->info,text);
  XMapWindow(display,window->info.id);
  XDisplayInfoString(display,&window->info,text);
  /*
    Load tile image.
  */
  XDefineCursor(display,window->image.id,window->image.busy_cursor);
  XFlush(display);
  (void) strcpy(resource_info->image_info->filename,filename);
  tiled_image=ReadImage(resource_info->image_info);
  XDefineCursor(display,window->image.id,window->image.cursor);
  XWithdrawWindow(display,window->info.id,window->info.screen);
  if (tiled_image == (Image *) NULL)
    XNoticeWidget(display,resource_info,window,"Unable to load image",filename);
  return(tiled_image);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X T r a n s l a t e I m a g e W i n d o w                                 %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XTranslateImageWindow translates the image within the image window
%  by one pixel as specified by the key symbol.  If the image has a `montage'
%  string the translation is respect to the width and height contained within
%  the string.
%
%  The format of the XTranslateImageWindow routine is:
%
%    XTranslateImageWindow(display,window,image,key_symbol)
%
%  A description of each parameter follows:
%
%    o display: Specifies a connection to an X server; returned from
%      XOpenDisplay.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o image: Specifies a pointer to a Image structure;  returned from
%      ReadImage.
%
%    o key_symbol: Specifies a KeySym which indicates which side of the image
%      to trim.
%
%
*/
static void XTranslateImageWindow(display,window,image,key_symbol)
Display
  *display;

XWindows
  *window;

Image
  *image;

KeySym
  key_symbol;
{
  int
    x,
    y;

  unsigned int
    x_offset,
    y_offset;

  /*
    User specified a pan position offset.
  */
  x_offset=1;
  y_offset=1;
  if (image->montage != (char *) NULL)
    (void) XParseGeometry(image->montage,&x,&y,&x_offset,&y_offset);
  switch (key_symbol)
  {
    case XK_Home:
    {
      window->image.x=window->image.width >> 1;
      window->image.y=window->image.height >> 1;
      break;
    }
    case XK_Left:
    {
      window->image.x-=x_offset;
      break;
    }
    case XK_Up:
    {
      window->image.y-=y_offset;
      break;
    }
    case XK_Right:
    {
      window->image.x+=x_offset;
      break;
    }
    case XK_Down:
    {
      window->image.y+=y_offset;
      break;
    }
    default:
      break;
  }
  /*
    Check boundary conditions.
  */
  if (window->image.x < 0)
    window->image.x=0;
  else
    if ((window->image.x+window->image.width) > window->image.ximage->width)
      window->image.x=window->image.ximage->width-window->image.width;
  if (window->image.y < 0)
    window->image.y=0;
  else
    if ((window->image.y+window->image.height) > window->image.ximage->height)
      window->image.y=window->image.ximage->height-window->image.height;
  /*
    Refresh image window.
  */
  XDrawPanRectangle(display,window);
  XRefreshWindow(display,&window->image,(XEvent *) NULL);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X W r i t e I m a g e W i n d o w                                         %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XWriteImageWindow writes an image to a file.
%
%  The format of the XWriteImageWindow routine is:
%
%    status=XWriteImageWindow(display,resource_info,window,image)
%
%  A description of each parameter follows:
%
%    o status: Function XWriteImageWindow return True if the image is
%      written.  False is returned is there is a memory shortage or if the
%      image fails to write.
%
%    o display: Specifies a connection to an X server; returned from
%      XOpenDisplay.
%
%    o resource_info: Specifies a pointer to a X11 XResourceInfo structure.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o image: Specifies a pointer to a Image structure;  returned from
%      ReadImage.
%
%
*/
static unsigned int XWriteImageWindow(display,resource_info,window,image)
Display
  *display;

XResourceInfo
  *resource_info;

XWindows
  *window;

Image
  *image;
{
  char
    filename[MaxTextLength],
    text[MaxTextLength];

  Image
    *output_image;

  ImageInfo
    *image_info;

  int
    status;

  /*
    Request file name from user.
  */
  (void) strcpy(filename,image->filename);
  if (resource_info->write_filename != (char *) NULL)
    (void) strcpy(filename,resource_info->write_filename);
  XFileBrowserWidget(display,resource_info,window,"Write",filename);
  if (*filename == '\0')
    return(True);
  image_info=resource_info->image_info;
  (void) strcpy(image_info->filename,filename);
  SetImageMagick(resource_info->image_info);
  if ((strcmp(image_info->magick,"JPEG") == 0) ||
      (strcmp(image_info->magick,"JPG") == 0))
    {
      char
        quality[MaxTextLength];

      /*
        Request JPEG quality from user.
      */
      (void) sprintf(quality,"%u",image_info->quality);
      XDialogWidget(display,resource_info,window,"Write","Enter JPEG quality:",
        quality);
      if (*quality != '\0')
        image_info->quality=atoi(quality);
    }
  if ((strcmp(image_info->magick,"EPS") == 0) ||
      (strcmp(image_info->magick,"PS") == 0) ||
      (strcmp(image_info->magick,"PS2") == 0))
    {
      char
        geometry[MaxTextLength];

      /*
        Request Postscript page geometry from user.
      */
      (void) sprintf(geometry,PSPageGeometry);
      if (image_info->page != (char *) NULL)
        (void) strcpy(geometry,image_info->page);
      XDialogWidget(display,resource_info,window,"Write",
        "Enter Postscript page geometry:",geometry);
      if (*geometry != '\0')
        {
          image_info->page=(char *) malloc((strlen(geometry)+1)*sizeof(char));
          if (image_info->page != (char *) NULL)
            (void) sprintf(image_info->page,geometry);
        }
    }
  /*
    Alert user we are busy.
  */
  XDefineCursor(display,window->image.id,window->image.busy_cursor);
  (void) strcpy(text," Writing image... ");
  XSetWindowExtents(display,&window->info,text);
  XMapWindow(display,window->info.id);
  XDisplayInfoString(display,&window->info,text);
  XFlush(display);
  /*
    Copy image before applying image transforms.
  */
  image->orphan=True;
  output_image=CopyImage(image,image->columns,image->rows,True);
  image->orphan=False;
  if (output_image == (Image *) NULL)
    {
      XDefineCursor(display,window->image.id,window->image.cursor);
      XWithdrawWindow(display,window->info.id,window->info.screen);
      return(False);
    }
  if ((window->image.clip_geometry != (char *) NULL) ||
     (output_image->columns != window->image.ximage->width) ||
     (output_image->rows != window->image.ximage->height))
    {
      char
        image_geometry[MaxTextLength];

      /*
        Clip and/or scale image.
      */
      (void) sprintf(image_geometry,"%dx%d",window->image.ximage->width,
        window->image.ximage->height);
      TransformImage(&output_image,window->image.clip_geometry,image_geometry);
    }
  if (resource_info->number_colors != 0)
    {
      /*
        Reduce the number of colors in the image.
      */
      if ((output_image->class == DirectClass) ||
          (output_image->colors > resource_info->number_colors) ||
          (resource_info->colorspace == GRAYColorspace))
        QuantizeImage(output_image,resource_info->number_colors,
          resource_info->tree_depth,resource_info->dither,
          resource_info->colorspace,True);
      SyncImage(output_image);
    }
  (void) strcpy(output_image->filename,filename);
  status=WriteImage(image_info,output_image);
  DestroyImage(output_image);
  XDefineCursor(display,window->image.id,window->image.cursor);
  XWithdrawWindow(display,window->info.id,window->info.screen);
  return(status);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%    M a i n                                                                  %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%
*/
int main(argc,argv)
int
  argc;

char
  **argv;
{
#define NotInitialized  (unsigned int) (~0)

  char
    *comment,
    *density,
    *label,
    *option,
    *page_geometry,
    *resource_value,
    *server_name;

  Display
    *display;

  Image
    *image,
    *next_image;

  ImageInfo
    image_info;

  int
    flags,
    i,
    p,
    q,
    x,
    y;

  unsigned int
    alpha,
    compression,
    first_scene,
    *image_marker,
    image_number,
    interlace,
    last_scene,
    quality,
    scene,
    verbose;

  unsigned long
    state;

  XrmDatabase
    resource_database,
    server_database;

  XResourceInfo
    resource_info;

  /*
    Set defaults.
  */
  alpha=NotInitialized;
  client_name=(*argv);
  display=(Display *) NULL;
  first_scene=0;
  image_marker=(unsigned int *) malloc(argc*sizeof(unsigned int));
  if (image_marker == (unsigned int *) NULL)
    Error("Unable to display image","Memory allocation failed");
  for (i=0; i < argc; i++)
    image_marker[i]=argc;
  image_number=0;
  last_scene=0;
  resource_database=(XrmDatabase) NULL;
  server_name=(char *) NULL;
  state=DefaultState;
  /*
    Check for server name specified on the command line.
  */
  for (i=1; i < argc; i++)
  {
    /*
      Check command line for server name.
    */
    option=argv[i];
    if (((int) strlen(option) > 1) && ((*option == '-') || (*option == '+')))
      if (strncmp("display",option+1,3) == 0)
        {
          /*
            User specified server name.
          */
          i++;
          if (i == argc)
            Error("Missing server name on -display",(char *) NULL);
          server_name=argv[i];
          break;
        }
  }
  /*
    Open X server connection.
  */
  display=XOpenDisplay(server_name);
  if (display == (Display *) NULL)
    Error("Unable to connect to X server",XDisplayName(server_name));
  /*
    Set our forgiving error handler.
  */
  XSetErrorHandler(XError);
  /*
    Initialize resource database.
  */
  XrmInitialize();
  XGetDefault(display,client_name,"dummy");
  resource_database=XrmGetDatabase(display);
  resource_value=XResourceManagerString(display);
  if (resource_value == (char *) NULL)
    resource_value="";
  server_database=XrmGetStringDatabase(resource_value);
  XrmMergeDatabases(server_database,&resource_database);
  /*
    Get user defaults from X resource database.
  */
  XGetResourceInfo(resource_database,client_name,&resource_info);
  resource_value=XGetResourceClass(resource_database,client_name,"compression",
    "RunlengthEncoded");
  if (Latin1Compare("qencoded",resource_value) == 0)
    compression=QEncodedCompression;
  else
    compression=RunlengthEncodedCompression;
  comment=
    XGetResourceClass(resource_database,client_name,"comment",(char *) NULL);
  density=
    XGetResourceClass(resource_database,client_name,"density",(char *) NULL);
  resource_value=
    XGetResourceClass(resource_database,client_name,"interlace","none");
  interlace=UndefinedInterlace;
  if (Latin1Compare("none",resource_value) == 0)
    interlace=NoneInterlace;
  if (Latin1Compare("line",resource_value) == 0)
    interlace=LineInterlace;
  if (Latin1Compare("plane",resource_value) == 0)
    interlace=PlaneInterlace;
  if (interlace == UndefinedInterlace)
    Warning("Unrecognized interlace type",resource_value);
  label=XGetResourceClass(resource_database,client_name,"label",(char *) NULL);
  page_geometry=XGetResourceClass(resource_database,client_name,"pageGeometry",
    (char *) NULL);
  resource_value=
    XGetResourceClass(resource_database,client_name,"quality","85");
  quality=atoi(resource_value);
  resource_value=XGetResourceClass(resource_database,client_name,"scene","0");
  scene=atoi(resource_value);
  resource_value=
    XGetResourceClass(resource_database,client_name,"verbose","False");
  verbose=IsTrue(resource_value);
  /*
    Parse command line.
  */
  p=1;
  for (i=1; ((i < argc) && !(state & ExitState)); i++)
  {
    option=argv[i];
    if (((int) strlen(option) > 1) && ((*option == '-') || (*option == '+')))
      switch (*(option+1))
      {
        case 'a':
        {
          alpha=(*option == '-');
          break;
        }
        case 'b':
        {
          if (strncmp("backdrop",option+1,5) == 0)
            {
              resource_info.backdrop=(*option == '-');
              break;
            }
          if (strncmp("background",option+1,5) == 0)
            {
              resource_info.background_color=(char *) NULL;
              if (*option == '-')
                {
                  i++;
                  if (i == argc)
                    Error("Missing color on -background",(char *) NULL);
                  resource_info.background_color=argv[i];
                }
              break;
            }
          if (strncmp("border",option+1,7) == 0)
            {
              if (*option == '-')
                {
                  i++;
                  if (i == argc)
                    Error("Missing geometry on -border",(char *) NULL);
                }
              break;
            }
          if (strncmp("bordercolor",option+1,7) == 0)
            {
              resource_info.border_color=(char *) NULL;
              if (*option == '-')
                {
                  i++;
                  if (i == argc)
                    Error("Missing color on -bordercolor",(char *) NULL);
                  resource_info.border_color=argv[i];
                }
              break;
            }
          if (strncmp("borderwidth",option+1,7) == 0)
            {
              resource_info.border_width=0;
              if (*option == '-')
                {
                  i++;
                  if ((i == argc) || !sscanf(argv[i],"%d",&x))
                    Error("Missing width on -borderwidth",(char *) NULL);
                  resource_info.border_width=atoi(argv[i]);
                }
              break;
            }
          Error("Unrecognized option",option);
          break;
        }
        case 'c':
        {
          if (strncmp("clip",option+1,2) == 0)
            {
              if (*option == '-')
                {
                  i++;
                  if ((i == argc) || !sscanf(argv[i],"%d",&x))
                    Error("Missing geometry on -clip",(char *) NULL);
                }
              break;
            }
          if (strncmp("colormap",option+1,6) == 0)
            {
              resource_info.colormap=PrivateColormap;
              if (*option == '-')
                {
                  i++;
                  if (i == argc)
                    Error("Missing type on -colormap",(char *) NULL);
                  option=argv[i];
                  resource_info.colormap=UndefinedColormap;
                  if (Latin1Compare("private",option) == 0)
                    resource_info.colormap=PrivateColormap;
                  if (Latin1Compare("shared",option) == 0)
                    resource_info.colormap=SharedColormap;
                  if (resource_info.colormap == UndefinedColormap)
                    Error("Invalid colormap type on -colormap",option);
                }
              break;
            }
          if (strncmp("colors",option+1,7) == 0)
            {
              resource_info.number_colors=0;
              if (*option == '-')
                {
                  i++;
                  if ((i == argc) || !sscanf(argv[i],"%d",&x))
                    Error("Missing colors on -colors",(char *) NULL);
                  resource_info.number_colors=atoi(argv[i]);
                }
              break;
            }
          if (strncmp("colorspace",option+1,7) == 0)
            {
              resource_info.colorspace=RGBColorspace;
              if (*option == '-')
                {
                  i++;
                  if (i == argc)
                    Error("Missing type on -colorspace",(char *) NULL);
                  option=argv[i];
                  resource_info.colorspace=UndefinedColorspace;
                  if (Latin1Compare("gray",option) == 0)
                    {
                      resource_info.colorspace=GRAYColorspace;
                      resource_info.number_colors=256;
                      resource_info.tree_depth=8;
                    }
                  if (Latin1Compare("ohta",option) == 0)
                    resource_info.colorspace=OHTAColorspace;
                  if (Latin1Compare("rgb",option) == 0)
                    resource_info.colorspace=RGBColorspace;
                  if (Latin1Compare("xyz",option) == 0)
                    resource_info.colorspace=XYZColorspace;
                  if (Latin1Compare("ycbcr",option) == 0)
                    resource_info.colorspace=YCbCrColorspace;
                  if (Latin1Compare("yiq",option) == 0)
                    resource_info.colorspace=YIQColorspace;
                  if (Latin1Compare("yuv",option) == 0)
                    resource_info.colorspace=YUVColorspace;
                  if (resource_info.colorspace == UndefinedColorspace)
                    Error("Invalid colorspace type on -colorspace",option);
                }
              break;
            }
          if (strncmp("comment",option+1,4) == 0)
            {
              comment=(char *) NULL;
              if (*option == '-')
                {
                  i++;
                  if (i == argc)
                    Error("Missing comment on -comment",(char *) NULL);
                  comment=argv[i];
                }
              break;
            }
          if (strncmp("compress",option+1,3) == 0)
            {
              compression=NoCompression;
              if (*option == '-')
                {
                  i++;
                  if (i == argc)
                    Error("Missing type on -compress",(char *) NULL);
                  option=argv[i];
                  if (Latin1Compare("runlengthencoded",option) == 0)
                    compression=RunlengthEncodedCompression;
                  else
                    if (Latin1Compare("qencoded",option) == 0)
                      compression=QEncodedCompression;
                    else
                      Error("Invalid compression type on -compress",option);
                }
              break;
            }
          Error("Unrecognized option",option);
          break;
        }
        case 'd':
        {
          if (strncmp("debug",option+1,3) == 0)
            {
              resource_info.debug=(*option == '-');
              break;
            }
          if (strncmp("delay",option+1,3) == 0)
            {
              resource_info.delay=0;
              if (*option == '-')
                {
                  i++;
                  if ((i == argc) || !sscanf(argv[i],"%d",&x))
                    Error("Missing seconds on -delay",(char *) NULL);
                  resource_info.delay=atoi(argv[i]);
                }
              break;
            }
          if (strncmp("density",option+1,3) == 0)
            {
              density=(char *) NULL;
              if (*option == '-')
                {
                  i++;
                  if ((i == argc) || !sscanf(argv[i],"%d",&x))
                    Error("Missing geometry on -density",(char *) NULL);
                  density=argv[i];
                }
              break;
            }
          if (strncmp("display",option+1,3) == 0)
            {
              server_name=(char *) NULL;
              if (*option == '-')
                {
                  i++;
                  if (i == argc)
                    Error("Missing server name on -display",(char *) NULL);
                  server_name=argv[i];
                }
              resource_info.server_name=server_name;
              break;
            }
          if (strncmp("dither",option+1,3) == 0)
            {
              resource_info.dither=(*option == '-');
              break;
            }
          Error("Unrecognized option",option);
          break;
        }
        case 'e':
        {
          if (strncmp("enhance",option+1,2) == 0)
            break;
          Error("Unrecognized option",option);
          break;
        }
        case 'f':
        {
          if (strncmp("font",option+1,3) == 0)
            {
              resource_info.font=(char *) NULL;
              if (*option == '-')
                {
                  i++;
                  if (i == argc)
                    Error("Missing font name on -font",(char *) NULL);
                  resource_info.font=argv[i];
                }
              break;
            }
         if (strncmp("foreground",option+1,3) == 0)
           {
             resource_info.foreground_color=(char *) NULL;
             if (*option == '-')
               {
                 i++;
                 if (i == argc)
                   Error("Missing foreground on -foreground",(char *) NULL);
                 resource_info.foreground_color=argv[i];
               }
              break;
           }
          Error("Unrecognized option",option);
          break;
        }
        case 'g':
        {
          if (strncmp("gamma",option+1,2) == 0)
            {
              if (*option == '-')
                {
                  i++;
                  if ((i == argc) || !sscanf(argv[i],"%f",(float *) &x))
                    Error("Missing value on -gamma",(char *) NULL);
                }
              break;
            }
          if (strncmp("geometry",option+1,2) == 0)
            {
              resource_info.image_geometry=(char *) NULL;
              if (*option == '-')
                {
                  i++;
                  if ((i == argc) || !sscanf(argv[i],"%d",&x))
                    Error("Missing geometry on -geometry",(char *) NULL);
                  resource_info.image_geometry=argv[i];
                }
              break;
            }
          Error("Unrecognized option",option);
          break;
        }
        case 'h':
        {
          if (strncmp("help",option+1,2) == 0)
            {
              Usage(True);
              break;
            }
          Error("Unrecognized option",option);
          break;
        }
        case 'i':
        {
          if (strncmp("iconGeometry",option+1,5) == 0)
            {
              resource_info.icon_geometry=(char *) NULL;
              if (*option == '-')
                {
                  i++;
                  if ((i == argc) || !sscanf(argv[i],"%d",&x))
                    Error("Missing geometry on -iconGeometry",(char *) NULL);
                  resource_info.icon_geometry=argv[i];
                }
              break;
            }
          if (strncmp("iconic",option+1,5) == 0)
            {
              resource_info.iconic=(*option == '-');
              break;
            }
          if (strncmp("interlace",option+1,3) == 0)
            {
              interlace=NoneInterlace;
              if (*option == '-')
                {
                  i++;
                  if (i == argc)
                    Error("Missing type on -interlace",(char *) NULL);
                  option=argv[i];
                  interlace=UndefinedInterlace;
                  if (Latin1Compare("none",option) == 0)
                    interlace=NoneInterlace;
                  if (Latin1Compare("line",option) == 0)
                    interlace=LineInterlace;
                  if (Latin1Compare("plane",option) == 0)
                    interlace=PlaneInterlace;
                  if (interlace == UndefinedInterlace)
                    Error("Invalid interlace type on -interlace",option);
                }
              break;
            }
          if (strncmp("inverse",option+1,2) == 0)
            break;
          Error("Unrecognized option",option);
          break;
        }
        case 'l':
        {
          if (strncmp("label",option+1,2) == 0)
            {
              label=(char *) NULL;
              if (*option == '-')
                {
                  i++;
                  if (i == argc)
                    Error("Missing label name on -label",(char *) NULL);
                  label=argv[i];
                }
              break;
            }
          Error("Unrecognized option",option);
          break;
        }
        case 'm':
        {
          if (strncmp("magnify",option+1,3) == 0)
            {
              resource_info.magnify=2;
              if (*option == '-')
                {
                  i++;
                  if ((i == argc) || !sscanf(argv[i],"%d",&x))
                    Error("Missing level on -magnify",(char *) NULL);
                  resource_info.magnify=atoi(argv[i]);
                }
              break;
            }
          if (strncmp("map",option+1,3) == 0)
            {
              resource_info.map_type=(char *) NULL;
              if (*option == '-')
                {
                  i++;
                  if (i == argc)
                    Error("Missing map type on -map",(char *) NULL);
                  resource_info.map_type=argv[i];
                }
              break;
            }
          if (strncmp("matte",option+1,2) == 0)
            {
              resource_info.matte_color=(char *) NULL;
              if (*option == '-')
                {
                  i++;
                  if (i == argc)
                    Error("Missing color on -matte",(char *) NULL);
                  resource_info.matte_color=argv[i];
                }
              break;
            }
          if (strncmp("monochrome",option+1,2) == 0)
            {
              resource_info.monochrome=(*option == '-');
              if (resource_info.monochrome)
                {
                  resource_info.number_colors=2;
                  resource_info.tree_depth=8;
                  resource_info.colorspace=GRAYColorspace;
                }
              break;
            }
          Error("Unrecognized option",option);
          break;
        }
        case 'n':
        {
          if (strncmp("name",option+1,2) == 0)
            {
              resource_info.name=(char *) NULL;
              if (*option == '-')
                {
                  i++;
                  if (i == argc)
                    Error("Missing name on -name",(char *) NULL);
                  resource_info.name=argv[i];
                }
              break;
            }
          if (strncmp("noise",option+1,3) == 0)
            break;
          if (strncmp("normalize",option+1,3) == 0)
            break;
          Error("Unrecognized option",option);
          break;
        }
        case 'p':
        {
          if (strncmp("page",option+1,2) == 0)
            {
              page_geometry=(char *) NULL;
              if (*option == '-')
                {
                  i++;
                  if ((i == argc) || !sscanf(argv[i],"%d",&x))
                    Error("Missing page geometry on -page",(char *) NULL);
                  page_geometry=argv[i];
                }
              break;
            }
          Error("Unrecognized option",option);
          break;
        }
        case 'q':
        {
          i++;
          if ((i == argc) || !sscanf(argv[i],"%d",&x))
            Error("Missing quality on -quality",(char *) NULL);
          quality=atoi(argv[i]);;
          break;
        }
        case 'r':
        {
          if (strncmp("reflect",option+1,2) == 0)
            break;
          if (strncmp("roll",option+1,3) == 0)
            {
              if (*option == '-')
                {
                  i++;
                  if ((i == argc) || !sscanf(argv[i],"%d",&x))
                    Error("Missing geometry on -roll",(char *) NULL);
                }
              break;
            }
          if (strncmp("rotate",option+1,3) == 0)
            {
              if (*option == '-')
                {
                  i++;
                  if ((i == argc) || !sscanf(argv[i],"%f",(float *) &x))
                    Error("Missing degrees on -rotate",(char *) NULL);
                }
              break;
            }
          Error("Unrecognized option",option);
          break;
        }
        case 's':
        {
          if (strncmp("scene",option+1,3) == 0)
            {
              first_scene=0;
              last_scene=0;
              if (*option == '-')
                {
                  i++;
                  if ((i == argc) || !sscanf(argv[i],"%d",&x))
                    Error("Missing scene number on -scene",(char *) NULL);
                  first_scene=atoi(argv[i]);
                  last_scene=first_scene;
                  (void) sscanf(argv[i],"%u-%u",&first_scene,&last_scene);
                }
              break;
            }
          if (strncmp("shear",option+1,2) == 0)
            {
              if (*option == '-')
                {
                  i++;
                  if ((i == argc) || !sscanf(argv[i],"%f",(float *) &x))
                    Error("Missing shear geometry on -shear",(char *) NULL);
                }
              break;
            }
          Error("Unrecognized option",option);
          break;
        }
        case 't':
        {
          if (strncmp("title",option+1,2) == 0)
            {
              resource_info.title=(char *) NULL;
              if (*option == '-')
                {
                  i++;
                  if (i == argc)
                    Error("Missing title on -title",(char *) NULL);
                  resource_info.title=argv[i];
                }
              break;
            }
          if (strncmp("treedepth",option+1,2) == 0)
            {
              resource_info.tree_depth=0;
              if (*option == '-')
                {
                  i++;
                  if ((i == argc) || !sscanf(argv[i],"%d",&x))
                    Error("Missing depth on -treedepth",(char *) NULL);
                  resource_info.tree_depth=atoi(argv[i]);
                }
              break;
            }
          Error("Unrecognized option",option);
          break;
        }
        case 'u':
        {
          if (strncmp("update",option+1,2) == 0)
            {
              resource_info.update=(*option == '-');
              if (*option == '-')
                {
                  i++;
                  if ((i == argc) || !sscanf(argv[i],"%d",&x))
                    Error("Missing seconds on -update",(char *) NULL);
                  resource_info.delay=atoi(argv[i]);
                }
              break;
            }
          if (strncmp("use_pixmap",option+1,2) == 0)
            {
              resource_info.use_pixmap=(*option == '-');
              break;
            }
          Error("Unrecognized option",option);
          break;
        }
        case 'v':
        {
          if (strncmp("verbose",option+1,2) == 0)
            {
              verbose=(*option == '-');
              break;
            }
          if (strncmp("visual",option+1,2) == 0)
            {
              resource_info.visual_type=(char *) NULL;
              if (*option == '-')
                {
                  i++;
                  if (i == argc)
                    Error("Missing visual class on -visual",(char *) NULL);
                  resource_info.visual_type=argv[i];
                }
              break;
            }
          Error("Unrecognized option",option);
          break;
        }
        case 'w':
        {
          if (strncmp("window",option+1,2) == 0)
            {
              resource_info.window_id=(char *) NULL;
              if (*option == '-')
                {
                  i++;
                  if (i == argc)
                    Error("Missing id, name, or 'root' on -window",
                      (char *) NULL);
                  resource_info.window_id=argv[i];
                }
              break;
            }
          if (strncmp("write",option+1,2) == 0)
            {
              resource_info.write_filename=(char *) NULL;
              if (*option == '-')
                {
                  i++;
                  if (i == argc)
                    Error("Missing file name on -write",(char *) NULL);
                  resource_info.write_filename=argv[i];
                  if (access(resource_info.write_filename,0) == 0)
                    {
                      char
                        answer[2];

                      (void) fprintf(stderr,"Overwrite %s? ",
                        resource_info.write_filename);
                      (void) gets(answer);
                      if (!((*answer == 'y') || (*answer == 'Y')))
                        exit(1);
                    }
                }
              break;
            }
          Error("Unrecognized option",option);
          break;
        }
        default:
        {
          Error("Unrecognized option",option);
          break;
        }
      }
    else
      for (scene=first_scene; scene <= last_scene ; scene++)
      {
        double
          normalized_maximum_error,
          normalized_mean_error;

        Image
          info_image;

        time_t
          start_time;

        unsigned int
          mean_error_per_pixel;

        unsigned long
          total_colors;

        /*
          Option is a file name: begin by reading image from specified file.
        */
        start_time=time((time_t *) NULL);
        GetImageInfo(argv[i],&image_info);
        if (first_scene != last_scene)
          {
            char
              filename[MaxTextLength];

            /*
              Form filename for multi-part images.
            */
            (void) sprintf(filename,image_info.filename,scene);
            if (strcmp(filename,image_info.filename) == 0)
              (void) sprintf(filename,"%s.%u",image_info.filename,scene);
            (void) strcpy(image_info.filename,filename);
          }
        image_info.server_name=resource_info.server_name;
        image_info.font=resource_info.font;
        image_info.geometry=resource_info.image_geometry;
        image_info.page=page_geometry;
        image_info.density=density;
        image_info.interlace=interlace;
        image_info.monochrome=resource_info.monochrome;
        image_info.quality=quality;
        image_info.verbose=verbose;
        resource_info.image_info=(&image_info);
        image=ReadImage(&image_info);
        if (image == (Image *) NULL)
          if (*option == '-')
            break;
          else
            continue;
        do
        {
          info_image=(*image);
          total_colors=0;
          if (alpha != NotInitialized)
            image->alpha=alpha;
          if (compression != UndefinedCompression)
            image->compression=compression;
          else
            image->compression=info_image.compression;
          if (scene != 0)
            image->scene=scene;
          LabelImage(image,label);
          if (comment != (char *) NULL)
            CommentImage(image,comment);
          /*
            Transform image as defined by the image geometries.
          */
          for (q=p; q < i; q++)
          {
            option=argv[q];
            if (strncmp("-border",option,8) == 0)
              {
                Image
                  *bordered_image;

                ImageInfo
                  border_info;

                /*
                  Surround image with a border of solid color.
                */
                q++;
                border_info=image_info;
                (void) strcpy(border_info.filename,"XC:");
                if (resource_info.border_color != (char *) NULL)
                  (void) strcat(border_info.filename,resource_info.border_color);
                border_info.geometry="1x1";
                bordered_image=ReadImage(&border_info);
                if (bordered_image != (Image *) NULL)
                  {
                    ColorPacket
                      border_color;

                    RectangleInfo
                      border_info;

                    DestroyImage(bordered_image);
                    border_info.width=0;
                    border_info.height=0;
                    border_info.x=0;
                    border_info.y=0;
                    flags=XParseGeometry(argv[q],&border_info.x,&border_info.y,
                      &border_info.width,&border_info.height);
                    if ((flags & HeightValue) == 0)
                      border_info.height=border_info.width;
                    border_color.red=bordered_image->pixels->red;
                    border_color.green=bordered_image->pixels->green;
                    border_color.blue=bordered_image->pixels->blue;
                    bordered_image=BorderImage(image,&border_info,&border_color,
                      &border_color);
                    if (bordered_image != (Image *) NULL)
                      {
                        DestroyImage(image);
                        bordered_image->class=DirectClass;
                        image=bordered_image;
                      }
                  }
              }
            if (strncmp("-clip",option,3) == 0)
              TransformImage(&image,argv[++q],(char *) NULL);
            if (strncmp("-enhance",option,2) == 0)
              {
                Image
                  *enhanced_image;

                /*
                  Enhance image.
                */
                enhanced_image=EnhanceImage(image);
                if (enhanced_image != (Image *) NULL)
                  {
                    DestroyImage(image);
                    image=enhanced_image;
                  }
              }
            if (strncmp("-gamma",option,3) == 0)
              GammaImage(image,argv[++q]);
            if (strncmp("-geometry",option,4) == 0)
              TransformImage(&image,(char *) NULL,argv[++q]);
            if (strncmp("-inverse",option,4) == 0)
              InverseImage(image);
            if (strncmp("-noise",option,4) == 0)
              {
                Image
                  *noisy_image;

                /*
                  Reduce noise in image.
                */
                noisy_image=NoisyImage(image);
                if (noisy_image != (Image *) NULL)
                  {
                    DestroyImage(image);
                    image=noisy_image;
                  }
              }
            if (strncmp("-normalize",option,4) == 0)
              NormalizeImage(image);
            if (strncmp("-reflect",option,3) == 0)
              {
                Image
                  *reflected_image;

                /*
                  Reverse image scanlines.
                */
                reflected_image=ReflectImage(image);
                if (reflected_image != (Image *) NULL)
                  {
                    DestroyImage(image);
                    image=reflected_image;
                  }
              }
            if (strncmp("-roll",option,4) == 0)
              {
                Image
                  *rolled_image;

                unsigned int
                  height,
                  width;

                /*
                  Roll image.
                */
                x=0;
                y=0;
                flags=XParseGeometry(argv[++q],&x,&y,&width,&height);
                rolled_image=RollImage(image,x,y);
                if (rolled_image != (Image *) NULL)
                  {
                    DestroyImage(image);
                    image=rolled_image;
                  }
              }
            if (strncmp("-rotate",option,4) == 0)
              {
                Image
                  *rotated_image;

                /*
                  Rotate image.
                */
                rotated_image=RotateImage(image,(double) atof(argv[++q]),False);
                if (rotated_image != (Image *) NULL)
                  {
                    DestroyImage(image);
                    image=rotated_image;
                  }
              }
            if (strncmp("-shear",option,3) == 0)
              {
                float
                  x_shear,
                  y_shear;

                Image
                  *sheared_image;

                /*
                  Shear image.
                */
                x_shear=0.0;
                y_shear=0.0;
                (void) sscanf(argv[++q],"%fx%f",&x_shear,&y_shear);
                sheared_image=ShearImage(image,x_shear,y_shear,False);
                if (sheared_image != (Image *) NULL)
                  {
                    DestroyImage(image);
                    image=sheared_image;
                  }
            }
          }
          if (resource_info.number_colors != 0)
            {
              /*
                Reduce the number of colors in the image.
              */
              if ((image->class == DirectClass) ||
                  (image->colors > resource_info.number_colors) ||
                  (resource_info.colorspace == GRAYColorspace))
                QuantizeImage(image,resource_info.number_colors,
                  resource_info.tree_depth,resource_info.dither,
                  resource_info.colorspace,True);
              if (verbose)
                {
                  /*
                    Measure quantization error.
                  */
                  QuantizationError(image,&mean_error_per_pixel,
                    &normalized_mean_error,&normalized_maximum_error);
                  total_colors=NumberColors(image,(FILE *) NULL);
                }
              SyncImage(image);
            }
          /*
            Display image to X server.
          */
          if (resource_info.window_id != (char *) NULL)
            {
              /*
                Display image to a specified X window.
              */
              XDisplayBackgroundImage(display,&resource_info,image);
              state&=ExitState;
            }
          else
            do
            {
              Image
                *loaded_image;

              /*
                Display montage image.
              */
              loaded_image=
                XDisplayImage(display,&resource_info,argv,argc,&image,&state);
              if (loaded_image == (Image *) NULL)
                break;
              while ((loaded_image != (Image *) NULL) && (!(state & ExitState)))
              {
                next_image=XDisplayImage(display,&resource_info,argv,argc,
                  &loaded_image,&state);
                DestroyImage(loaded_image);
                loaded_image=next_image;
              }
            } while (!(state & ExitState));
          if (resource_info.write_filename != (char *) NULL)
            {
              /*
                Write image.
              */
              (void) strcpy(image->filename,resource_info.write_filename);
              (void) WriteImage(&image_info,image);
            }
          if (verbose)
            {
              /*
                Display detailed info about the image.
              */
              (void) fprintf(stderr,"[%u] %s",
                image->scene == 0 ? image_number : image->scene,
                info_image.filename);
              if (resource_info.write_filename != (char *) NULL)
                (void) fprintf(stderr,"=>%s",resource_info.write_filename);
              (void) fprintf(stderr," %ux%u",info_image.columns,
                info_image.rows);
              if ((info_image.columns != image->columns) ||
                  (info_image.rows != image->rows))
                (void) fprintf(stderr,"=>%ux%u",image->columns,image->rows);
              if (image->class == DirectClass)
                (void) fprintf(stderr," DirectClass ");
              else
                if (total_colors == 0)
                  (void) fprintf(stderr," PseudoClass %uc",image->colors);
                else
                  {
                    (void) fprintf(stderr," PseudoClass %lu=>%uc",total_colors,
                      image->colors);
                    (void) fprintf(stderr," %u/%.6f/%.6fe",mean_error_per_pixel,
                      normalized_mean_error,normalized_maximum_error);
                  }
              (void) fprintf(stderr," %s %lds\n",image->magick,
                time((time_t *) NULL)-start_time+1);
            }
          /*
            Proceed to next/previous image.
          */
          if (state & FormerImageState)
            next_image=image->previous;
          else
            next_image=image->next;
          if (next_image != (Image *) NULL)
            image=next_image;
        } while ((next_image != (Image *) NULL) && !(state & ExitState));
        /*
          Free image resources.
        */
        while (image->previous != (Image *) NULL)
          image=image->previous;
        DestroyImages(image);
        if (!(state & FormerImageState))
          image_marker[i]=image_number++;
        else
          {
            /*
              Proceed to previous image.
            */
            for (i--; i > 0; i--)
              if (image_marker[i] == (image_number-2))
                break;
            if (image_number != 0)
              image_number--;
          }
        p=i+1;
        if (state & ExitState)
          break;
      }
    if (i == (argc-1))
      if ((resource_info.delay != 0) && (image_number != 1))
        {
          /*
            Proceed to first image.
          */
          i=0;
          image_number=0;
        }
  }
  if (image_number == 0)
    {
      Image
        *loaded_image;

      /*
        Display background image.
      */
      GetImageInfo("xc:",&image_info);
      image_info.server_name=resource_info.server_name;
      image_info.font=resource_info.font;
      image_info.geometry=resource_info.image_geometry;
      image_info.page=page_geometry;
      image_info.density=density;
      image_info.interlace=interlace;
      image_info.monochrome=resource_info.monochrome;
      image_info.quality=quality;
      image_info.verbose=verbose;
      resource_info.image_info=(&image_info);
      image=ReadImage(&image_info);
      loaded_image=
        XDisplayImage(display,&resource_info,argv,argc,&image,&state);
      DestroyImage(image);
      image=loaded_image;
      if (loaded_image != (Image *) NULL)
        do
        {
          loaded_image=
            XDisplayImage(display,&resource_info,argv,argc,&image,&state);
          if (loaded_image == (Image *) NULL)
            break;
          while ((loaded_image != (Image *) NULL) && (!(state & ExitState)))
          {
            next_image=XDisplayImage(display,&resource_info,argv,argc,
              &loaded_image,&state);
            DestroyImage(loaded_image);
            loaded_image=next_image;
          }
        } while (!(state & ExitState));
    }
  XCloseDisplay(display);
  (void) free((char *) image_marker);
  return(False);
}
